class S3UploadError(Exception):
    def __init__(self, reason: str):
        self.reason = reason

    def __str__(self) -> str:
        return f"S3 file upload failed, reason: {self.reason}"


class S3UploadFileError(S3UploadError):
    def __init__(
        self,
        file: str,
        reason: str,
    ):
        super().__init__(reason)
        self.file = file

    def __str__(self) -> str:
        return f"S3 upload failed for file: '{self.file}'. Reason: {self.reason}"


class S3UploadErrorFileIsTooLarge(S3UploadFileError):
    def __init__(self, file: str, reason: str):
        super().__init__(file, reason)


class S3UploadErrorFileIsEmpty(S3UploadFileError):
    def __init__(self, file: str, reason: str):
        super().__init__(file, reason)



import httpcore
import functools
import contextlib

from typing import Iterator, Callable


@contextlib.contextmanager
def async_http_connections_expire_immediately() -> Iterator[None]:
    """
    This patching addresses the issue of httpx.AsyncClient not working
    correctly when it's used by multiple event loops.

    The connection from connection pool created with one event loop can be tried to be used
    by the request processed via another event loop. Asyncio doesn't support
    that and the RuntimeError is raised.

    So, this context manager patches AsyncHTTPConnection class in a way that all of the
    async connections expire immediately and the runtime error is not possible.

    Related issues:
    https://github.com/comet-ml/opik/issues/1132
    https://github.com/encode/httpx/discussions/2959

    TODO: this function might probably require extra logic for handling the cases
    when there is already existing async connection pool with opened connections, but it is
    out of scope for now.
    """
    try:
        original = httpcore.AsyncHTTPConnection.__init__

        def AsyncHTTPConnection__init__wrapper() -> Callable:
            @functools.wraps(original)
            def wrapped(*args, **kwargs):  # type: ignore
                kwargs["keepalive_expiry"] = 0
                return original(*args, **kwargs)

            return wrapped

        httpcore.AsyncHTTPConnection.__init__ = AsyncHTTPConnection__init__wrapper()
        yield
    finally:
        httpcore.AsyncHTTPConnection.__init__ = original



import logging
import re
from collections.abc import Mapping
from typing import Any, Dict, Literal, Optional, cast

from opik import _logging

LOGGER = logging.getLogger(__name__)

SpanType = Literal["llm", "tool", "general"]


def get_span_type(run: Dict[str, Any]) -> SpanType:
    if run.get("run_type") in ["llm", "tool"]:
        return cast(SpanType, run.get("run_type"))

    if run.get("run_type") in ["prompt"]:
        return cast(SpanType, "tool")

    return cast(SpanType, "general")


def is_root_run(run_dict: Dict[str, Any]) -> bool:
    return run_dict.get("parent_run_id") is None


def extract_tool_description(run_dict: Dict[str, Any]) -> Optional[str]:
    """
    Extract tool description/docstring from a tool Run object.

    The function looks for the description in the following locations (in order):
    1. serialized["kwargs"]["description"] - explicit description attribute
    2. serialized["description"] - top-level description in serialized data

    Args:
        run_dict: Dictionary representation of a LangChain Run object.

    Returns:
        The tool description as a string if found, None otherwise.
    """
    serialized = run_dict.get("serialized")
    if not isinstance(serialized, dict):
        return None

    # Try to get description from kwargs (most common location for @tool decorated functions)
    kwargs = serialized.get("kwargs", {})
    if isinstance(kwargs, dict):
        description = kwargs.get("description")
        if description and isinstance(description, str):
            return description

    # Try top-level description in serialized data
    description = serialized.get("description")
    if description and isinstance(description, str):
        return description

    return None


def get_run_metadata(run_dict: Dict[str, Any]) -> Dict[str, Any]:
    extra = run_dict.get("extra") or {}
    if not isinstance(extra, dict):
        extra = {}
    metadata = extra.get("metadata", {}).copy()

    # Extract tool description for tool runs and add to metadata
    if run_dict.get("run_type") == "tool":
        tool_description = extract_tool_description(run_dict)
        if tool_description:
            metadata["tool_description"] = tool_description

    return metadata


def parse_graph_interrupt_value(error_traceback: str) -> Optional[str]:
    """
    Parse GraphInterrupt error traceback to extract the interrupt value as a string.

    The function extracts the value from the Interrupt object representation in the traceback.
    It handles both string values (with quotes) and non-string values, including nested structures.
    For string values, escape sequences are decoded (e.g., \\n becomes a newline character).

    Args:
        error_traceback: The error traceback string containing GraphInterrupt information.

    Returns:
        The interrupt value as a string if found, None otherwise.
    """
    # Search for GraphInterrupt( anywhere in the traceback
    match = re.search(
        r"GraphInterrupt\(.*?Interrupt\(value=",
        error_traceback,
        re.DOTALL,
    )
    if not match:
        return None

    # Start parsing from after "value="
    start_pos = match.end()
    value_str = error_traceback[start_pos:]

    # Extract the value, handling nested parentheses and brackets
    paren_depth = 0
    bracket_depth = 0
    brace_depth = 0
    in_string = False
    string_char = None
    i = 0

    for i, char in enumerate(value_str):
        # Handle string boundaries
        if char in ('"', "'") and (i == 0 or value_str[i - 1] != "\\"):
            if not in_string:
                in_string = True
                string_char = char
            elif char == string_char:
                in_string = False
                string_char = None

        # Skip counting brackets/parens inside strings
        if in_string:
            continue

        # Track nesting depth
        if char == "(":
            paren_depth += 1
        elif char == ")":
            if paren_depth > 0:
                paren_depth -= 1
            else:
                # Found the closing paren of Interrupt(...), stop here
                break
        elif char == "[":
            bracket_depth += 1
        elif char == "]":
            bracket_depth -= 1
        elif char == "{":
            brace_depth += 1
        elif char == "}":
            brace_depth -= 1
        elif (
            char == "," and paren_depth == 0 and bracket_depth == 0 and brace_depth == 0
        ):
            # Found a comma at the top level, stop here
            break

    # Extract and clean the value
    value = value_str[:i].strip()

    # Check if the value was originally a quoted string
    was_quoted_string = False
    if len(value) >= 2 and value[0] in ('"', "'") and value[-1] == value[0]:
        was_quoted_string = True
        value = value[1:-1]

    # Decode escape sequences for string values
    if was_quoted_string:
        try:
            value = value.encode("utf-8").decode("unicode_escape")
        except (UnicodeDecodeError, AttributeError):
            # If decoding fails, return the original value
            pass

    return value


def extract_resume_value_from_command(obj: Any) -> Optional[str]:
    """
    Extract the resume value from a LangGraph Command object or serialized Command dict.

    Args:
        obj: A Command object or dict representing a serialized Command object (from run.dict()).

    Returns:
        The resume value as a string if found, None otherwise.
    """
    # Check if it's a Command object (has a resume attribute)
    if hasattr(obj, "resume") and obj.resume is not None:
        return str(obj.resume)
    # Check if it's a serialized Command dict
    if obj is not None and isinstance(obj, dict) and "resume" in obj:
        return str(obj["resume"])
    return None


def extract_command_update(outputs: Dict[str, Any]) -> Dict[str, Any]:
    """Extract state updates from LangGraph Command objects.

    When a LangGraph node returns a Command, LangChain wraps it in {"output": Command(...)}.
    This function detects Command objects and extracts the update dict to properly log state changes.

    Args:
        outputs: The outputs dict from a LangChain Run.

    Returns:
        The extracted update dict if a Command is found, otherwise the original outputs.
    """
    if "output" in outputs and len(outputs) == 1:
        output_value = outputs["output"]
        # Duck-type check for Command object
        if hasattr(output_value, "update") and hasattr(output_value, "goto"):
            try:
                update_value = output_value.update
                # Handle None - skip extraction
                if update_value is None:
                    return outputs
                # Accept any Mapping type and convert to dict for JSON serialization
                if isinstance(update_value, Mapping):
                    _logging.log_once_at_level(
                        logging.DEBUG,
                        "Extracted state update from LangGraph Command object",
                        LOGGER,
                    )
                    return dict(update_value)
            except Exception as e:
                LOGGER.warning(
                    f"Failed to extract update from Command-like object: {e}",
                    exc_info=True,
                )

    return outputs



FEEDBACK_SCORE_SOURCE_SDK = "sdk"
DATASET_SOURCE_SDK = "sdk"

FEEDBACK_SCORES_MAX_BATCH_SIZE = 1000
EXPERIMENT_ITEMS_MAX_BATCH_SIZE = 1000
DATASET_ITEMS_MAX_BATCH_SIZE = 1000
DELETE_TRACE_BATCH_SIZE = 1000

DATASET_STREAM_BATCH_SIZE = 2000



import logging
from typing import Optional, NamedTuple, List, Literal, cast

from opik.api_objects.attachment import (
    attachments_extractor,
    attachment_context,
    converters,
)

from . import message_processors
from ..preprocessing import constants
from .. import messages, streamer


LOGGER = logging.getLogger(__name__)


class EntityDetails(NamedTuple):
    entity_type: Literal["span", "trace"]
    entity_id: str
    project_name: str


class AttachmentsExtractionProcessor(message_processors.BaseMessageProcessor):
    """
    Class for processing message attachments through extraction and further handling.

    The AttachmentsExtractionProcessor class is designed to handle attachments from incoming
    messages. It checks the type of messages and processes them if they support
    attachments. This includes extracting attachment data, replacing them with references,
    and streaming processed or original messages through a pipeline. The class provides a
    mechanism to toggle processing activity and ensures proper handling of messages with
    embedded attachment information.
    """

    def __init__(
        self,
        min_attachment_size: int,
        messages_streamer: streamer.Streamer,
        url_override: str,
        is_active: bool = True,
    ):
        """
        Initializes an object with essential components for managing message streaming
        and attachment extraction.

        Args:
            min_attachment_size: Minimum size for an attachment to be extracted.
            messages_streamer: The streamer that is responsible for managing
                messages broadcasts.
            url_override: A custom URL to override default configurations if set.
            is_active: Indicator of whether this instance is active. Default is True.
        """
        self._is_active = is_active
        self.extractor = attachments_extractor.AttachmentsExtractor(min_attachment_size)
        self.messages_streamer = messages_streamer
        self._url_override = url_override

        self.attachment_attributes = ["input", "output", "metadata"]

    def is_active(self) -> bool:
        return self._is_active

    def process(self, message: messages.BaseMessage) -> None:
        if not isinstance(message, messages.AttachmentSupportingMessage):
            return

        if self._is_active:
            # do attachment processing only if the processor is active
            try:
                self._process_attachments_in_message(message.original_message)
            except Exception as ex:
                LOGGER.error(
                    "Failed to process attachment support message: %s", ex, exc_info=ex
                )

        # put the original message into the streamer for further processing
        original_message = message.original_message
        setattr(original_message, constants.MARKER_ATTRIBUTE_NAME, True)
        self.messages_streamer.put(original_message)

    def _process_attachments_in_message(self, original: messages.BaseMessage) -> None:
        entity_details = entity_type_from_attachment_message(original)
        if entity_details is None:
            LOGGER.error(
                "Failed to extract entity details from message - %s. Skipping embedded attachments processing.",
                original.__class__.__name__,
            )
            return

        attachments = []

        for attribute in self.attachment_attributes:
            if getattr(original, attribute, None):
                results = self.extractor.extract_and_replace(
                    data=getattr(original, attribute),
                    entity_type=entity_details.entity_type,
                    entity_id=entity_details.entity_id,
                    project_name=entity_details.project_name,
                    context=cast(Literal["input", "output", "metadata"], attribute),
                )
                attachments.extend(results)

        if len(attachments) > 0:
            LOGGER.debug(
                "Extracted %d attachments from %s (entity: %s/%s)",
                len(attachments),
                original.__class__.__name__,
                entity_details.entity_type,
                entity_details.entity_id,
            )

            self._process_attachments(attachments)
        else:
            LOGGER.debug(
                "No attachments found in the message - %s.", original.__class__.__name__
            )

    def _process_attachments(
        self, attachments: List[attachment_context.AttachmentWithContext]
    ) -> None:
        for attachment in attachments:
            create_attachment_message = converters.attachment_to_message(
                attachment_data=attachment.attachment_data,
                entity_type=attachment.entity_type,
                entity_id=attachment.entity_id,
                project_name=attachment.project_name,
                url_override=self._url_override,
                delete_after_upload=True,  # make sure to delete attachments after upload to avoid leaking space and data
            )
            self.messages_streamer.put(create_attachment_message)


def entity_type_from_attachment_message(
    message: messages.BaseMessage,
) -> Optional[EntityDetails]:
    if isinstance(message, (messages.CreateSpanMessage, messages.UpdateSpanMessage)):
        return EntityDetails("span", message.span_id, project_name=message.project_name)
    elif isinstance(
        message, (messages.CreateTraceMessage, messages.UpdateTraceMessage)
    ):
        return EntityDetails(
            "trace", message.trace_id, project_name=message.project_name
        )
    else:
        return None



from typing import Any, Dict, List, Optional, Tuple
import json
import dataclasses

import opik.exceptions
from opik.rest_api import client as rest_client
from opik.rest_api import core as rest_api_core
from opik.rest_api.types import prompt_version_detail
from . import types as prompt_types


@dataclasses.dataclass
class PromptSearchResult:
    """Result from searching prompts, containing name, template structure, and latest version details."""

    name: str
    template_structure: str
    prompt_version_detail: prompt_version_detail.PromptVersionDetail


class PromptClient:
    def __init__(self, client: rest_client.OpikApi):
        self._rest_client = client

    def create_prompt(
        self,
        name: str,
        prompt: str,
        metadata: Optional[Dict[str, Any]],
        type: prompt_types.PromptType = prompt_types.PromptType.MUSTACHE,
        template_structure: str = "text",
        id: Optional[str] = None,
        description: Optional[str] = None,
        change_description: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ) -> prompt_version_detail.PromptVersionDetail:
        """
        Creates the prompt detail for the given prompt name and template.

        Parameters:
        - name: The name of the prompt.
        - prompt: The template content for the prompt.
        - metadata: Optional metadata for the prompt.
        - type: The template type (MUSTACHE or JINJA2).
        - template_structure: Either "text" (default) or "chat".
        - id: Optional unique identifier (UUID) for the prompt.
        - description: Optional description of the prompt (up to 255 characters).
        - change_description: Optional description of changes in this version.
        - tags: Optional list of tags to associate with the prompt.

        Returns:
        - A Prompt object for the provided prompt name and template.

        Raises:
        - PromptTemplateStructureMismatch: If a prompt with the same name already exists but has a different
          template_structure (e.g., trying to create a text prompt when a chat prompt exists, or vice versa).
          Template structure is immutable after prompt creation.
        """
        prompt_version = self._get_latest_version(name)

        # For chat prompts, compare parsed JSON to avoid formatting differences
        templates_equal = False

        if prompt_version is not None:
            if prompt_version.template_structure != template_structure:
                raise opik.exceptions.PromptTemplateStructureMismatch(
                    prompt_name=name,
                    existing_structure=prompt_version.template_structure,
                    attempted_structure=template_structure,
                )

            if template_structure == "chat":
                try:
                    existing_messages = json.loads(prompt_version.template)
                    new_messages = json.loads(prompt)
                    templates_equal = existing_messages == new_messages
                except (json.JSONDecodeError, TypeError):
                    templates_equal = prompt_version.template == prompt
            else:
                templates_equal = prompt_version.template == prompt

        # Create a new version if:
        # - No version exists yet (new prompt)
        # - Template content has changed
        # - Metadata has changed
        # - Type has changed
        # Note: template_structure is immutable and used by the backend only if it is the first prompt version.
        if (
            prompt_version is None
            or not templates_equal
            or prompt_version.metadata != metadata
            or prompt_version.type != type.value
        ):
            prompt_version = self._create_new_version(
                name=name,
                prompt=prompt,
                type=type,
                metadata=metadata,
                template_structure=template_structure,
                id=id,
                description=description,
                change_description=change_description,
                tags=tags,
            )

        return prompt_version

    def _create_new_version(
        self,
        name: str,
        prompt: str,
        type: prompt_version_detail.PromptVersionDetailType,
        metadata: Optional[Dict[str, Any]],
        template_structure: str = "text",
        id: Optional[str] = None,
        description: Optional[str] = None,
        change_description: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ) -> prompt_version_detail.PromptVersionDetail:
        # Check if this is a new prompt (no existing versions)
        existing_version = self._get_latest_version(name)

        # If it's a new prompt and container-level params are provided, use create_prompt endpoint
        # which creates both the container and first version in one call
        if existing_version is None and (
            id is not None or description is not None or tags is not None
        ):
            self._rest_client.prompts.create_prompt(
                name=name,
                id=id,
                description=description,
                template=prompt,
                metadata=metadata,
                change_description=change_description,
                type=type,
                template_structure=template_structure,
                tags=tags,
            )
            # After creating, retrieve the version that was created
            new_prompt_version_detail = (
                self._rest_client.prompts.retrieve_prompt_version(
                    name=name,
                )
            )
            # retrieve_prompt_version may not return tags, so we need to set them manually
            # from the tags we just passed to create_prompt
            if tags is not None and new_prompt_version_detail.tags is None:
                # Pydantic objects are frozen, so we need to create a new object with tags
                new_prompt_version_detail = prompt_version_detail.PromptVersionDetail(
                    id=new_prompt_version_detail.id,
                    prompt_id=new_prompt_version_detail.prompt_id,
                    commit=new_prompt_version_detail.commit,
                    template=new_prompt_version_detail.template,
                    metadata=new_prompt_version_detail.metadata,
                    type=new_prompt_version_detail.type,
                    change_description=new_prompt_version_detail.change_description,
                    tags=tags,
                    variables=new_prompt_version_detail.variables,
                    template_structure=new_prompt_version_detail.template_structure,
                    created_at=new_prompt_version_detail.created_at,
                    created_by=new_prompt_version_detail.created_by,
                )
        else:
            # For existing prompts or when no container-level params, use create_prompt_version
            new_prompt_version_detail_data = prompt_version_detail.PromptVersionDetail(
                template=prompt,
                metadata=metadata,
                type=type,
            )
            new_prompt_version_detail = self._rest_client.prompts.create_prompt_version(
                name=name,
                version=new_prompt_version_detail_data,
                template_structure=template_structure,
            )
        return new_prompt_version_detail

    def _get_latest_version(
        self, name: str
    ) -> Optional[prompt_version_detail.PromptVersionDetail]:
        return self.get_prompt(name=name, commit=None)

    def get_prompt(
        self,
        name: str,
        commit: Optional[str] = None,
        raise_if_not_template_structure: Optional[str] = None,
    ) -> Optional[prompt_version_detail.PromptVersionDetail]:
        """
        Retrieve the prompt detail for a given prompt name and commit version.

        Parameters:
            name: The name of the prompt.
            commit: An optional commit version of the prompt. If not provided, the latest version is retrieved.
            raise_if_not_template_structure: Optional template structure validation. If provided and doesn't match, raises PromptTemplateStructureMismatch.

        Returns:
            Prompt: The details of the specified prompt.
        """
        try:
            prompt_version = self._rest_client.prompts.retrieve_prompt_version(
                name=name,
                commit=commit,
            )

            should_skip_validation = (
                prompt_version.template_structure is None
                and raise_if_not_template_structure == "text"
            )
            if should_skip_validation:
                return prompt_version

            # Client-side validation for template_structure if requested and not skipped
            if (
                raise_if_not_template_structure is not None
                and prompt_version.template_structure != raise_if_not_template_structure
            ):
                raise opik.exceptions.PromptTemplateStructureMismatch(
                    prompt_name=name,
                    existing_structure=prompt_version.template_structure,
                    attempted_structure=raise_if_not_template_structure,
                )

            return prompt_version
        except rest_api_core.ApiError as e:
            if e.status_code != 404:
                raise e
            # 400, 404 - not found
        return None

    # TODO: Need to add support for prompt name in the BE so we don't
    # need to retrieve the prompt id
    def get_all_prompt_versions(
        self, name: str
    ) -> List[prompt_version_detail.PromptVersionDetail]:
        """
        Retrieve all the prompt details for a given prompt name.

        Parameters:
            name: The name of the prompt.

        Returns:
            List[Prompt]: A list of prompts for the given name.
        """
        try:
            prompts_matching_name_string = self._rest_client.prompts.get_prompts(
                name=name
            ).content
            if (
                prompts_matching_name_string is None
                or len(prompts_matching_name_string) == 0
            ):
                raise ValueError("No prompts found for name: " + name)

            filtered_prompt_list = [
                x.id for x in prompts_matching_name_string if name == x.name
            ]
            if len(filtered_prompt_list) == 0:
                raise ValueError("No prompts found for name: " + name)

            prompt_id = filtered_prompt_list[0]
            return self._get_prompt_versions_by_id_paginated(prompt_id)

        except rest_api_core.ApiError as e:
            if e.status_code != 404:
                raise e

        return []

    def _get_prompt_versions_by_id_paginated(
        self, prompt_id: str
    ) -> List[prompt_version_detail.PromptVersionDetail]:
        page = 1
        size = 100
        prompts: List[prompt_version_detail.PromptVersionDetail] = []
        while True:
            prompt_versions_page = self._rest_client.prompts.get_prompt_versions(
                id=prompt_id, page=page, size=size
            ).content

            versions = prompt_versions_page or []
            prompts.extend(
                [
                    # Converting to PromptVersionDetail for consistency with other methods.
                    # TODO: backend should implement non-frontend endpoint which will return PromptVersionDetail objects
                    prompt_version_detail.PromptVersionDetail(
                        id=version.id,
                        prompt_id=version.prompt_id,
                        template=version.template,
                        type=version.type,
                        metadata=version.metadata,
                        commit=version.commit,
                        created_at=version.created_at,
                        created_by=version.created_by,
                    )
                    for version in versions
                ]
            )

            if len(versions) < size:
                break
            page += 1

        return prompts

    def search_prompts(
        self,
        *,
        name: Optional[str] = None,
        parsed_filters: Optional[List[Dict[str, Any]]] = None,
    ) -> List[PromptSearchResult]:
        """
        Search prompt containers by optional name substring and filters, then
        return the latest version detail for each matched prompt container.

        Parameters:
            name: Optional substring of the prompt name to search for.
            parsed_filters: List of parsed filters (OQL) that will be stringified for the backend.

        Returns:
            List[PromptSearchResult]: Each result contains name, template_structure, and prompt_version_detail.
        """
        try:
            filters_str = (
                json.dumps(parsed_filters) if parsed_filters is not None else None
            )

            # Page through all prompt containers and collect:
            # (name, template_structure, tags)
            page = 1
            size = 1000
            prompt_info: List[Tuple[str, str, Optional[List[str]]]] = []
            while True:
                prompts_page = self._rest_client.prompts.get_prompts(
                    page=page,
                    size=size,
                    name=name,
                    filters=filters_str,
                )
                content = prompts_page.content or []
                if len(content) == 0:
                    break
                prompt_info.extend(
                    [(p.name, p.template_structure or "text", p.tags) for p in content]
                )
                if len(content) < size:
                    break
                page += 1

            if len(prompt_info) == 0:
                return []

            # Retrieve latest version for each container name
            results: List[PromptSearchResult] = []
            for prompt_name, template_structure, tags in prompt_info:
                try:
                    latest_version = self._rest_client.prompts.retrieve_prompt_version(
                        name=prompt_name,
                    )
                    # retrieve_prompt_version may not return tags, so we need to set them from get_prompts response
                    if tags is not None and latest_version.tags is None:
                        # Pydantic objects are frozen, so we need to create a new object with tags
                        latest_version = prompt_version_detail.PromptVersionDetail(
                            id=latest_version.id,
                            prompt_id=latest_version.prompt_id,
                            commit=latest_version.commit,
                            template=latest_version.template,
                            metadata=latest_version.metadata,
                            type=latest_version.type,
                            change_description=latest_version.change_description,
                            tags=tags,
                            variables=latest_version.variables,
                            template_structure=latest_version.template_structure,
                            created_at=latest_version.created_at,
                            created_by=latest_version.created_by,
                        )
                    results.append(
                        PromptSearchResult(
                            name=prompt_name,
                            template_structure=template_structure,
                            prompt_version_detail=latest_version,
                        )
                    )
                except rest_api_core.ApiError as e:
                    # Skip prompts that can't be retrieved (e.g., deleted between search and retrieval)
                    if e.status_code == 404:
                        continue
                    raise e

            return results

        except rest_api_core.ApiError as e:
            if e.status_code != 404:
                raise e
            return []



# This file was auto-generated by Fern from our API Definition.

import typing

import httpx
from .alerts.client import AlertsClient, AsyncAlertsClient
from .annotation_queues.client import AnnotationQueuesClient, AsyncAnnotationQueuesClient
from .attachments.client import AsyncAttachmentsClient, AttachmentsClient
from .automation_rule_evaluators.client import AsyncAutomationRuleEvaluatorsClient, AutomationRuleEvaluatorsClient
from .chat_completions.client import AsyncChatCompletionsClient, ChatCompletionsClient
from .check.client import AsyncCheckClient, CheckClient
from .core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from .core.request_options import RequestOptions
from .dashboards.client import AsyncDashboardsClient, DashboardsClient
from .datasets.client import AsyncDatasetsClient, DatasetsClient
from .environment import OpikApiEnvironment
from .experiments.client import AsyncExperimentsClient, ExperimentsClient
from .feedback_definitions.client import AsyncFeedbackDefinitionsClient, FeedbackDefinitionsClient
from .guardrails.client import AsyncGuardrailsClient, GuardrailsClient
from .llm_provider_key.client import AsyncLlmProviderKeyClient, LlmProviderKeyClient
from .manual_evaluation.client import AsyncManualEvaluationClient, ManualEvaluationClient
from .ollama.client import AsyncOllamaClient, OllamaClient
from .open_telemetry_ingestion.client import AsyncOpenTelemetryIngestionClient, OpenTelemetryIngestionClient
from .optimizations.client import AsyncOptimizationsClient, OptimizationsClient
from .projects.client import AsyncProjectsClient, ProjectsClient
from .prompts.client import AsyncPromptsClient, PromptsClient
from .raw_client import AsyncRawOpikApi, RawOpikApi
from .redirect.client import AsyncRedirectClient, RedirectClient
from .service_toggles.client import AsyncServiceTogglesClient, ServiceTogglesClient
from .spans.client import AsyncSpansClient, SpansClient
from .system_usage.client import AsyncSystemUsageClient, SystemUsageClient
from .traces.client import AsyncTracesClient, TracesClient
from .welcome_wizard.client import AsyncWelcomeWizardClient, WelcomeWizardClient
from .workspaces.client import AsyncWorkspacesClient, WorkspacesClient


class OpikApi:
    """
    Use this class to access the different functions within the SDK. You can instantiate any number of clients with different configuration that will propagate to these functions.

    Parameters
    ----------
    base_url : typing.Optional[str]
        The base url to use for requests from the client.

    environment : OpikApiEnvironment
        The environment to use for requests from the client. from .environment import OpikApiEnvironment

        Defaults to OpikApiEnvironment.DEFAULT



    api_key : typing.Optional[str]
    workspace_name : typing.Optional[str]
    timeout : typing.Optional[float]
        The timeout to be used, in seconds, for requests. By default the timeout is 60 seconds, unless a custom httpx client is used, in which case this default is not enforced.

    follow_redirects : typing.Optional[bool]
        Whether the default httpx client follows redirects or not, this is irrelevant if a custom httpx client is passed in.

    httpx_client : typing.Optional[httpx.Client]
        The httpx client to use for making requests, a preconfigured client is used by default, however this is useful should you want to pass in any custom httpx configuration.

    Examples
    --------
    from Opik import OpikApi
    client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
    """

    def __init__(
        self,
        *,
        base_url: typing.Optional[str] = None,
        environment: OpikApiEnvironment = OpikApiEnvironment.DEFAULT,
        api_key: typing.Optional[str] = None,
        workspace_name: typing.Optional[str] = None,
        timeout: typing.Optional[float] = None,
        follow_redirects: typing.Optional[bool] = True,
        httpx_client: typing.Optional[httpx.Client] = None,
    ):
        _defaulted_timeout = (
            timeout if timeout is not None else 60 if httpx_client is None else httpx_client.timeout.read
        )
        self._client_wrapper = SyncClientWrapper(
            base_url=_get_base_url(base_url=base_url, environment=environment),
            api_key=api_key,
            workspace_name=workspace_name,
            httpx_client=httpx_client
            if httpx_client is not None
            else httpx.Client(timeout=_defaulted_timeout, follow_redirects=follow_redirects)
            if follow_redirects is not None
            else httpx.Client(timeout=_defaulted_timeout),
            timeout=_defaulted_timeout,
        )
        self._raw_client = RawOpikApi(client_wrapper=self._client_wrapper)
        self.system_usage = SystemUsageClient(client_wrapper=self._client_wrapper)
        self.alerts = AlertsClient(client_wrapper=self._client_wrapper)
        self.annotation_queues = AnnotationQueuesClient(client_wrapper=self._client_wrapper)
        self.attachments = AttachmentsClient(client_wrapper=self._client_wrapper)
        self.check = CheckClient(client_wrapper=self._client_wrapper)
        self.automation_rule_evaluators = AutomationRuleEvaluatorsClient(client_wrapper=self._client_wrapper)
        self.chat_completions = ChatCompletionsClient(client_wrapper=self._client_wrapper)
        self.dashboards = DashboardsClient(client_wrapper=self._client_wrapper)
        self.datasets = DatasetsClient(client_wrapper=self._client_wrapper)
        self.experiments = ExperimentsClient(client_wrapper=self._client_wrapper)
        self.feedback_definitions = FeedbackDefinitionsClient(client_wrapper=self._client_wrapper)
        self.guardrails = GuardrailsClient(client_wrapper=self._client_wrapper)
        self.llm_provider_key = LlmProviderKeyClient(client_wrapper=self._client_wrapper)
        self.manual_evaluation = ManualEvaluationClient(client_wrapper=self._client_wrapper)
        self.ollama = OllamaClient(client_wrapper=self._client_wrapper)
        self.open_telemetry_ingestion = OpenTelemetryIngestionClient(client_wrapper=self._client_wrapper)
        self.optimizations = OptimizationsClient(client_wrapper=self._client_wrapper)
        self.projects = ProjectsClient(client_wrapper=self._client_wrapper)
        self.prompts = PromptsClient(client_wrapper=self._client_wrapper)
        self.service_toggles = ServiceTogglesClient(client_wrapper=self._client_wrapper)
        self.spans = SpansClient(client_wrapper=self._client_wrapper)
        self.traces = TracesClient(client_wrapper=self._client_wrapper)
        self.welcome_wizard = WelcomeWizardClient(client_wrapper=self._client_wrapper)
        self.workspaces = WorkspacesClient(client_wrapper=self._client_wrapper)
        self.redirect = RedirectClient(client_wrapper=self._client_wrapper)

    @property
    def with_raw_response(self) -> RawOpikApi:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawOpikApi
        """
        return self._raw_client

    def is_alive(self, *, request_options: typing.Optional[RequestOptions] = None) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            default response

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.is_alive()
        """
        _response = self._raw_client.is_alive(request_options=request_options)
        return _response.data

    def version(self, *, request_options: typing.Optional[RequestOptions] = None) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            default response

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.version()
        """
        _response = self._raw_client.version(request_options=request_options)
        return _response.data


class AsyncOpikApi:
    """
    Use this class to access the different functions within the SDK. You can instantiate any number of clients with different configuration that will propagate to these functions.

    Parameters
    ----------
    base_url : typing.Optional[str]
        The base url to use for requests from the client.

    environment : OpikApiEnvironment
        The environment to use for requests from the client. from .environment import OpikApiEnvironment

        Defaults to OpikApiEnvironment.DEFAULT



    api_key : typing.Optional[str]
    workspace_name : typing.Optional[str]
    timeout : typing.Optional[float]
        The timeout to be used, in seconds, for requests. By default the timeout is 60 seconds, unless a custom httpx client is used, in which case this default is not enforced.

    follow_redirects : typing.Optional[bool]
        Whether the default httpx client follows redirects or not, this is irrelevant if a custom httpx client is passed in.

    httpx_client : typing.Optional[httpx.AsyncClient]
        The httpx client to use for making requests, a preconfigured client is used by default, however this is useful should you want to pass in any custom httpx configuration.

    Examples
    --------
    from Opik import AsyncOpikApi
    client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
    """

    def __init__(
        self,
        *,
        base_url: typing.Optional[str] = None,
        environment: OpikApiEnvironment = OpikApiEnvironment.DEFAULT,
        api_key: typing.Optional[str] = None,
        workspace_name: typing.Optional[str] = None,
        timeout: typing.Optional[float] = None,
        follow_redirects: typing.Optional[bool] = True,
        httpx_client: typing.Optional[httpx.AsyncClient] = None,
    ):
        _defaulted_timeout = (
            timeout if timeout is not None else 60 if httpx_client is None else httpx_client.timeout.read
        )
        self._client_wrapper = AsyncClientWrapper(
            base_url=_get_base_url(base_url=base_url, environment=environment),
            api_key=api_key,
            workspace_name=workspace_name,
            httpx_client=httpx_client
            if httpx_client is not None
            else httpx.AsyncClient(timeout=_defaulted_timeout, follow_redirects=follow_redirects)
            if follow_redirects is not None
            else httpx.AsyncClient(timeout=_defaulted_timeout),
            timeout=_defaulted_timeout,
        )
        self._raw_client = AsyncRawOpikApi(client_wrapper=self._client_wrapper)
        self.system_usage = AsyncSystemUsageClient(client_wrapper=self._client_wrapper)
        self.alerts = AsyncAlertsClient(client_wrapper=self._client_wrapper)
        self.annotation_queues = AsyncAnnotationQueuesClient(client_wrapper=self._client_wrapper)
        self.attachments = AsyncAttachmentsClient(client_wrapper=self._client_wrapper)
        self.check = AsyncCheckClient(client_wrapper=self._client_wrapper)
        self.automation_rule_evaluators = AsyncAutomationRuleEvaluatorsClient(client_wrapper=self._client_wrapper)
        self.chat_completions = AsyncChatCompletionsClient(client_wrapper=self._client_wrapper)
        self.dashboards = AsyncDashboardsClient(client_wrapper=self._client_wrapper)
        self.datasets = AsyncDatasetsClient(client_wrapper=self._client_wrapper)
        self.experiments = AsyncExperimentsClient(client_wrapper=self._client_wrapper)
        self.feedback_definitions = AsyncFeedbackDefinitionsClient(client_wrapper=self._client_wrapper)
        self.guardrails = AsyncGuardrailsClient(client_wrapper=self._client_wrapper)
        self.llm_provider_key = AsyncLlmProviderKeyClient(client_wrapper=self._client_wrapper)
        self.manual_evaluation = AsyncManualEvaluationClient(client_wrapper=self._client_wrapper)
        self.ollama = AsyncOllamaClient(client_wrapper=self._client_wrapper)
        self.open_telemetry_ingestion = AsyncOpenTelemetryIngestionClient(client_wrapper=self._client_wrapper)
        self.optimizations = AsyncOptimizationsClient(client_wrapper=self._client_wrapper)
        self.projects = AsyncProjectsClient(client_wrapper=self._client_wrapper)
        self.prompts = AsyncPromptsClient(client_wrapper=self._client_wrapper)
        self.service_toggles = AsyncServiceTogglesClient(client_wrapper=self._client_wrapper)
        self.spans = AsyncSpansClient(client_wrapper=self._client_wrapper)
        self.traces = AsyncTracesClient(client_wrapper=self._client_wrapper)
        self.welcome_wizard = AsyncWelcomeWizardClient(client_wrapper=self._client_wrapper)
        self.workspaces = AsyncWorkspacesClient(client_wrapper=self._client_wrapper)
        self.redirect = AsyncRedirectClient(client_wrapper=self._client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawOpikApi:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawOpikApi
        """
        return self._raw_client

    async def is_alive(self, *, request_options: typing.Optional[RequestOptions] = None) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            default response

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.is_alive()
        asyncio.run(main())
        """
        _response = await self._raw_client.is_alive(request_options=request_options)
        return _response.data

    async def version(self, *, request_options: typing.Optional[RequestOptions] = None) -> typing.Optional[typing.Any]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.Optional[typing.Any]
            default response

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.version()
        asyncio.run(main())
        """
        _response = await self._raw_client.version(request_options=request_options)
        return _response.data


def _get_base_url(*, base_url: typing.Optional[str] = None, environment: OpikApiEnvironment) -> str:
    if base_url is not None:
        return base_url
    elif environment is not None:
        return environment.value
    else:
        raise Exception("Please pass in either base_url or environment to construct the client")



import base64
import logging
import os
import shutil
import tempfile
from typing import Literal, Optional

from ...file_upload import mime_type
from ...message_processing import messages
from . import attachment

LOGGER = logging.getLogger(__name__)


def attachment_to_message(
    attachment_data: attachment.Attachment,
    entity_type: Literal["trace", "span"],
    entity_id: str,
    project_name: str,
    url_override: str,
    delete_after_upload: bool = False,
) -> messages.CreateAttachmentMessage:
    if attachment_data.data is None:
        raise ValueError("Attachment data cannot be None")

    mimetype = guess_attachment_type(attachment_data)
    base_url_path = base64.b64encode(url_override.encode("utf-8")).decode("utf-8")

    if isinstance(attachment_data.data, bytes):
        file_path = _write_file_like_to_temp_file(attachment_data.data)
        # make sure the temporary file is deleted after upload
        attachment_data.create_temp_copy = False
        delete_after_upload = True
    else:
        file_path = attachment_data.data

    file_name = attachment_data.file_name
    should_delete_after_upload = delete_after_upload

    if file_name is None:
        file_name = os.path.basename(file_path)

    # Try to create a temporary copy if requested
    if attachment_data.create_temp_copy:
        tmp_file_path = _try_create_temp_copy(file_path)
        if tmp_file_path is not None:
            file_path = tmp_file_path
            should_delete_after_upload = True
        else:
            should_delete_after_upload = False

    return messages.CreateAttachmentMessage(
        file_path=file_path,
        file_name=file_name,
        mime_type=mimetype,
        entity_type=entity_type,
        entity_id=entity_id,
        project_name=project_name,
        encoded_url_override=base_url_path,
        delete_after_upload=should_delete_after_upload,
    )


def _write_file_like_to_temp_file(file_like: bytes) -> str:
    """
    Writes a bytes-like file object to a temporary file on the filesystem.

    This function accepts a file-like object in the form of bytes and writes its
    contents to a temporary file. The temporary file is not deleted automatically
    on program termination. The path to the created temporary file is returned.

    Args:
        file_like: The bytes-like object containing the file content
            to be written to a temporary file.

    Returns:
        The full path to the created temporary file.
    """
    temp_file = tempfile.NamedTemporaryFile(mode="wb", delete=False)
    try:
        temp_file.write(file_like)
        temp_file.flush()
        LOGGER.debug(
            "Created temporary copy of file-like attachment to file: %s",
            temp_file.name,
        )
    except Exception as e:
        LOGGER.error(
            "Failed to write file-like attachment to temp file: %s",
            e,
            exc_info=True,
        )
        raise
    finally:
        temp_file.close()

    return temp_file.name


def _try_create_temp_copy(file_path: str) -> Optional[str]:
    """
    Create a temporary copy of a file.

    This ensures the file remains available for upload even if the user
    deletes the original file. The temp file is created with delete=False
    so it persists until the upload manager processes and deletes it.

    Args:
        file_path: Path to the original file.

    Returns:
        Path to the temporary copy.
    """
    _, extension = os.path.splitext(file_path)
    temp_file = tempfile.NamedTemporaryFile(mode="wb", delete=False, suffix=extension)
    try:
        with open(file_path, "rb") as original_file:
            shutil.copyfileobj(original_file, temp_file)
        temp_file.flush()
        temp_file.close()
        LOGGER.debug(
            "Created temporary copy of attachment: %s -> %s",
            file_path,
            temp_file.name,
        )
        return temp_file.name
    except Exception:
        temp_file.close()
        LOGGER.error(
            "Failed to create temporary copy of attachment: %s. Opik will try to use the original file.",
            file_path,
            exc_info=True,
        )
        return None


def guess_attachment_type(attachment_data: attachment.Attachment) -> Optional[str]:
    if attachment_data.content_type is not None:
        return attachment_data.content_type

    mimetype = None
    if attachment_data.file_name is not None:
        mimetype = mime_type.guess_mime_type(file=attachment_data.file_name)

    if mimetype is not None:
        return mimetype

    if isinstance(attachment_data.data, str):
        mimetype = mime_type.guess_mime_type(file=attachment_data.data)
    elif isinstance(attachment_data.data, bytes):
        mimetype = mime_type.BINARY_MIME_TYPE

    return mimetype



"""
Decorator for google.genai.Client.operations.get method.

When the operation returns a completed video, patches the Video.save method
to track video saves with attachments.
"""

import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
)

from typing_extensions import override

from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator
from opik.decorator import inspect_helpers

from . import video_save_decorator

if TYPE_CHECKING:
    from google.genai.types import GenerateVideosOperation

LOGGER = logging.getLogger(__name__)


class OperationsGetTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    Decorator for tracking Google GenAI operations.get method.

    When the operation is done and contains videos, patches Video.save
    to track video saves with attachments.
    """

    def __init__(self, project_name: Optional[str], upload_videos: bool) -> None:
        super().__init__()
        self._project_name = project_name
        self._upload_videos = upload_videos

    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> arguments_helpers.StartSpanParameters:
        name = track_options.name if track_options.name is not None else func.__name__
        input_data = inspect_helpers.extract_inputs(func, args, kwargs)

        return arguments_helpers.StartSpanParameters(
            name=name,
            input=input_data,
            type=track_options.type,
            tags=track_options.tags,
            metadata=track_options.metadata,
            project_name=track_options.project_name,
        )

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Optional["GenerateVideosOperation"],
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        if output is not None and output.done and output.response:
            video_save_decorator.patch_videos_save(
                output,
                project_name=self._project_name,
                tags=current_span_data.tags,
                metadata=current_span_data.metadata,
                upload_video=self._upload_videos,
            )

        return arguments_helpers.EndSpanParameters(output=output)

    @override
    def _streams_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], Any]],
    ) -> Optional[Any]:
        NOT_A_STREAM = None
        return NOT_A_STREAM



from .flow import patch_flow
from .llm_client import patch_llm_client
from .litellm_completion import patch_litellm_completion

__all__ = ["patch_flow", "patch_llm_client", "patch_litellm_completion"]



from typing import Optional, TypeVar

import openai
import opik

from . import (
    chat_completion_chunks_aggregator,
    openai_chat_completions_decorator,
)
import opik.semantic_version as semantic_version

OpenAIClient = TypeVar("OpenAIClient", openai.OpenAI, openai.AsyncOpenAI)


def _get_provider(openai_client: OpenAIClient) -> str:
    """Get the provider name from the OpenAI client's base URL."""
    if openai_client.base_url.host != "api.openai.com":
        return openai_client.base_url.host
    return "openai"


def track_openai(
    openai_client: OpenAIClient,
    project_name: Optional[str] = None,
) -> OpenAIClient:
    """Adds Opik tracking wrappers to an OpenAI client.

    The client is always patched; however every wrapped call checks
    `opik.is_tracing_active()` before emitting
    any telemetry. If tracing is disabled at call time, the wrapped function
    executes normally but no span/trace is sent.

    Tracks calls to:
    * `openai_client.chat.completions.create()`, including support for stream=True mode.
    * `openai_client.beta.chat.completions.parse()`
    * `openai_client.beta.chat.completions.stream()`
    * `openai_client.responses.create()`
    * `openai_client.videos.create()`, `videos.create_and_poll()`, `videos.poll()`,
      `videos.list()`, `videos.delete()`, `videos.remix()`, `videos.download_content()`,
      and `write_to_file()` on downloaded content

    Can be used within other Opik-tracked functions.

    Args:
        openai_client: An instance of OpenAI or AsyncOpenAI client.
        project_name: The name of the project to log data.

    Returns:
        The modified OpenAI client with Opik tracking enabled.
    """
    if hasattr(openai_client, "opik_tracked"):
        return openai_client

    openai_client.opik_tracked = True

    _patch_openai_chat_completions(openai_client, project_name)

    if hasattr(openai_client, "responses"):
        _patch_openai_responses(openai_client, project_name)

    if hasattr(openai_client, "videos"):
        _patch_openai_videos(openai_client, project_name)

    return openai_client


def _patch_openai_chat_completions(
    openai_client: OpenAIClient,
    project_name: Optional[str] = None,
) -> None:
    chat_completions_decorator_factory = (
        openai_chat_completions_decorator.OpenaiChatCompletionsTrackDecorator()
    )
    chat_completions_decorator_factory.provider = _get_provider(openai_client)

    completions_create_decorator = chat_completions_decorator_factory.track(
        type="llm",
        name="chat_completion_create",
        generations_aggregator=chat_completion_chunks_aggregator.aggregate,
        project_name=project_name,
    )
    completions_parse_decorator = chat_completions_decorator_factory.track(
        type="llm",
        name="chat_completion_parse",
        generations_aggregator=chat_completion_chunks_aggregator.aggregate,
        project_name=project_name,
    )
    completions_stream_decorator = chat_completions_decorator_factory.track(
        type="llm",
        name="chat_completion_stream",
        generations_aggregator=chat_completion_chunks_aggregator.aggregate,
        project_name=project_name,
    )

    openai_client.chat.completions.create = completions_create_decorator(
        openai_client.chat.completions.create
    )
    if semantic_version.SemanticVersion.parse(openai.__version__) < "1.92.0":  # type: ignore
        # beta.chat.completions.stream() calls chat.completions.create(stream=True)
        # under the hood.
        # So decorating `create` will automatically work for tracking `stream`.
        openai_client.beta.chat.completions.parse = completions_parse_decorator(
            openai_client.beta.chat.completions.parse
        )
    else:
        # OpenAI reworked beta API.
        # * chat.completion.stream calls chat.completion.create under the hood, so
        #   it doesn't need to be decorated again.
        # * But beta.chat.completion.stream does not call chat.completion.create, so
        #   it needs to be decorated!
        openai_client.beta.chat.completions.stream = completions_stream_decorator(
            openai_client.beta.chat.completions.stream
        )

        openai_client.chat.completions.parse = completions_parse_decorator(
            openai_client.chat.completions.parse
        )
        openai_client.beta.chat.completions.parse = completions_parse_decorator(
            openai_client.beta.chat.completions.parse
        )


def _patch_openai_responses(
    openai_client: OpenAIClient,
    project_name: Optional[str] = None,
) -> None:
    from . import (
        response_events_aggregator,
        openai_responses_decorator,
    )

    responses_decorator_factory = (
        openai_responses_decorator.OpenaiResponsesTrackDecorator()
    )
    responses_decorator_factory.provider = _get_provider(openai_client)

    if hasattr(openai_client.responses, "create"):
        responses_create_decorator = responses_decorator_factory.track(
            type="llm",
            name="responses_create",
            generations_aggregator=response_events_aggregator.aggregate,
            project_name=project_name,
        )
        openai_client.responses.create = responses_create_decorator(
            openai_client.responses.create
        )

    if hasattr(openai_client.responses, "parse"):
        responses_parse_decorator = responses_decorator_factory.track(
            type="llm",
            name="responses_parse",
            generations_aggregator=response_events_aggregator.aggregate,
            project_name=project_name,
        )
        openai_client.responses.parse = responses_parse_decorator(
            openai_client.responses.parse
        )


def _patch_openai_videos(
    openai_client: OpenAIClient,
    project_name: Optional[str] = None,
) -> None:
    from . import videos

    provider = _get_provider(openai_client)
    create_decorator_factory = videos.VideosCreateTrackDecorator(provider=provider)
    download_decorator_factory = videos.VideosDownloadTrackDecorator()

    video_metadata = {"created_from": "openai", "type": "openai_videos"}
    video_tags = ["openai"]

    if hasattr(openai_client.videos, "create"):
        decorator = create_decorator_factory.track(
            type="llm",
            name="videos.create",
            project_name=project_name,
        )
        openai_client.videos.create = decorator(openai_client.videos.create)

    if hasattr(openai_client.videos, "create_and_poll"):
        decorator = opik.track(
            name="videos.create_and_poll",
            tags=video_tags,
            metadata=video_metadata,
            project_name=project_name,
        )
        openai_client.videos.create_and_poll = decorator(
            openai_client.videos.create_and_poll
        )

    if hasattr(openai_client.videos, "remix"):
        decorator = create_decorator_factory.track(
            type="llm",
            name="videos.remix",
            project_name=project_name,
        )
        openai_client.videos.remix = decorator(openai_client.videos.remix)

    # Note: videos.retrieve is intentionally NOT patched to avoid too many spans
    # since it's called frequently during polling operations.

    if hasattr(openai_client.videos, "poll"):
        decorator = opik.track(
            name="videos.poll",
            tags=video_tags,
            metadata=video_metadata,
            project_name=project_name,
        )
        openai_client.videos.poll = decorator(openai_client.videos.poll)

    if hasattr(openai_client.videos, "delete"):
        decorator = opik.track(
            name="videos.delete",
            tags=video_tags,
            metadata=video_metadata,
            project_name=project_name,
        )
        openai_client.videos.delete = decorator(openai_client.videos.delete)

    # Patch download_content - also patches write_to_file on returned instances
    # download_content returns a lazy response object, write_to_file does the actual download
    if hasattr(openai_client.videos, "download_content"):
        decorator = download_decorator_factory.track(
            type="general",
            name="videos.download_content",
            project_name=project_name,
        )
        openai_client.videos.download_content = decorator(
            openai_client.videos.download_content
        )

    if hasattr(openai_client.videos, "list"):
        decorator = opik.track(
            name="videos.list",
            tags=video_tags,
            metadata=video_metadata,
            project_name=project_name,
        )
        openai_client.videos.list = decorator(openai_client.videos.list)



from typing import Union, List, Dict, Callable, Tuple

from . import anonymizer, rules_anonymizer, rules

RulesType = Union[
    List[Dict[str, str]],
    List[Tuple[str, str]],
    List[Callable[[str], str]],
    List[Union[Dict[str, str], Tuple[str, str], Callable[[str], str]]],
    Dict[str, str],
    Tuple[str, str],
    Callable[[str], str],
]


def create_anonymizer(
    anonymizer_rules: RulesType, max_depth: int = 10
) -> anonymizer.Anonymizer:
    """Create an anonymizer with the specified rules.

    Args:
        anonymizer_rules: Anonymizer rules specification in various formats:
            - Dict with "regex" and "replace" keys for a single regex rule
            - Tuple with (regex, replacement) for a single regex rule
            - Callable that takes a string and returns anonymized string
            - List of any of the above for multiple rules
        max_depth: Maximum recursion depth for nested data structures.

    Returns:
        An Anonymizer instance configured with the specified rules.

    Raises:
        ValueError: If a rule format is invalid.
    """
    rule_objects: List[rules.Rule] = []

    if callable(anonymizer_rules):
        # Single function rule
        rule_objects.append(rules.FunctionRule(anonymizer_rules))
    elif isinstance(anonymizer_rules, dict):
        # Single dictionary rule
        _check_dictionary_rule(anonymizer_rules)
        rule_objects.append(
            rules.RegexRule(anonymizer_rules["regex"], anonymizer_rules["replace"])
        )
    elif isinstance(anonymizer_rules, tuple):
        # Single tuple rule
        _check_tuple_rule(anonymizer_rules)
        regex_pattern, replacement = anonymizer_rules
        rule_objects.append(rules.RegexRule(regex_pattern, replacement))
    elif isinstance(anonymizer_rules, list):
        # List of rules
        for rule in anonymizer_rules:
            if callable(rule) and not isinstance(rule, (dict, tuple)):
                rule_objects.append(rules.FunctionRule(rule))
            elif isinstance(rule, dict):
                _check_dictionary_rule(rule)
                rule_objects.append(rules.RegexRule(rule["regex"], rule["replace"]))
            elif isinstance(rule, tuple):
                _check_tuple_rule(rule)
                regex_pattern, replacement = rule
                rule_objects.append(rules.RegexRule(regex_pattern, replacement))
            else:
                raise ValueError(f"Unsupported rule type in list: {type(rule)}")
    else:
        raise ValueError(f"Unsupported rules type: {type(anonymizer_rules)}")

    return rules_anonymizer.RulesAnonymizer(rule_objects, max_depth=max_depth)


def _check_dictionary_rule(rule: Dict[str, str]) -> None:
    if "regex" not in rule or "replace" not in rule:
        raise ValueError("Dictionary rule must have 'regex' and 'replace' keys")


def _check_tuple_rule(rule: Tuple[str, str]) -> None:
    if len(rule) != 2:
        raise ValueError(
            "Tuple rule must have exactly 2 elements: (regex, replacement)"
        )



import logging
from typing import List, Optional

from opik.api_objects import dataset, experiment, opik_client
from opik.types import BatchFeedbackScoreDict
from . import test_case
from .metrics import score_result
from .types import ScoringKeyMappingType

LOGGER = logging.getLogger(__name__)


def get_experiment_with_unique_name(
    client: opik_client.Opik, experiment_name: str
) -> experiment.Experiment:
    experiments = client.get_experiments_by_name(name=experiment_name)

    if len(experiments) == 0:
        raise ValueError(f"Experiment with name {experiment_name} not found")
    elif len(experiments) > 1:
        raise ValueError(
            f"Found multiple experiments with name {experiment_name}. Try to use `experiment_id` instead"
        )

    experiment_ = experiments[0]
    return experiment_


def get_trace_project_name(client: opik_client.Opik, trace_id: str) -> str:
    # We first need to get the project_id for the trace
    traces = client.get_trace_content(id=trace_id)
    project_id = traces.project_id

    # Then we can get the project name
    project_metadata = client.get_project(id=project_id)
    return project_metadata.name


def get_experiment_test_cases(
    experiment_: experiment.Experiment,
    dataset_: dataset.Dataset,
    scoring_key_mapping: Optional[ScoringKeyMappingType],
) -> List[test_case.TestCase]:
    experiment_items = experiment_.get_items()

    # Fetch dataset items to get input data for bulk-uploaded experiment items
    dataset_items_by_id = {item["id"]: item for item in dataset_.get_items()}

    test_cases = []
    for item in experiment_items:
        dataset_item_data = dataset_items_by_id.get(item.dataset_item_id)

        if dataset_item_data is None:
            LOGGER.error(
                f"Unexpected error: Dataset item with id {item.dataset_item_id} not found, skipping experiment item {item.id}"
            )
            continue

        if item.evaluation_task_output is None:
            LOGGER.error(
                f"Unexpected error: Evaluation task output is None for experiment item {item.id}, skipping experiment item"
            )
            continue

        test_cases.append(
            test_case.TestCase(
                trace_id=item.trace_id,
                dataset_item_id=item.dataset_item_id,
                task_output=item.evaluation_task_output,
                dataset_item_content=dataset_item_data,
            )
        )

    return test_cases


def log_test_result_feedback_scores(
    client: opik_client.Opik,
    score_results: List[score_result.ScoreResult],
    trace_id: str,
    project_name: Optional[str],
) -> None:
    all_trace_scores: List[BatchFeedbackScoreDict] = []

    for score_result_ in score_results:
        if score_result_.scoring_failed:
            continue

        trace_score = BatchFeedbackScoreDict(
            id=trace_id,
            name=score_result_.name,
            value=score_result_.value,
            reason=score_result_.reason,
        )
        all_trace_scores.append(trace_score)

    if len(all_trace_scores) > 0:
        client.log_traces_feedback_scores(
            scores=all_trace_scores, project_name=project_name
        )



import math
import os
from typing import IO, Optional

from .s3_upload_error import S3UploadErrorFileIsEmpty, S3UploadErrorFileIsTooLarge

# Constants defining existing AWS S3 limits (https://docs.aws.amazon.com/AmazonS3/latest/userguide/qfacts.html)
MAX_SUPPORTED_PARTS_NUMBER = 10000
MIN_FILE_PART_SIZE = 5 * 1024 * 1024
MAX_FILE_PART_SIZE = 5 * 1024 * 1024 * 1024


class BaseFilePartsStrategy:
    """Describes strategy to be used to split file into multiple parts for AWS S3 upload.

    Args:
        file: the path to the file to be uploaded.
        file_size: the size of the file to be uploaded.
        max_file_part_size: the maximum size of the file to be uploaded (default: MIN_FILE_PART_SIZE).
    """

    def __init__(
        self,
        file: str,
        file_size: int,
        max_file_part_size: int = MIN_FILE_PART_SIZE,
    ):
        self.file = file
        self.file_size = file_size

        if max_file_part_size < MIN_FILE_PART_SIZE:
            self.max_file_part_size = MIN_FILE_PART_SIZE
        elif max_file_part_size > MAX_FILE_PART_SIZE:
            self.max_file_part_size = MAX_FILE_PART_SIZE
        else:
            self.max_file_part_size = max_file_part_size

    def calculate(self) -> int:
        if self.file_size == 0:
            raise S3UploadErrorFileIsEmpty(self.file, "Can not upload empty file")

        parts_number = math.ceil(self.file_size / self.max_file_part_size)

        if parts_number > MAX_SUPPORTED_PARTS_NUMBER:
            parts_number = MAX_SUPPORTED_PARTS_NUMBER

        # check that we are still in part size limits
        part_size = math.ceil(self.file_size / parts_number)
        if part_size > MAX_FILE_PART_SIZE:
            raise S3UploadErrorFileIsTooLarge(
                self.file,
                "File is too large to be uploaded, file size: %d" % self.file_size,
            )

        if part_size < MIN_FILE_PART_SIZE:
            self.max_file_part_size = MIN_FILE_PART_SIZE
        else:
            self.max_file_part_size = part_size

        return parts_number


class FilePartsStrategy(BaseFilePartsStrategy):
    def __init__(
        self,
        file_path: str,
        file_size: Optional[int] = None,
        max_file_part_size: int = MIN_FILE_PART_SIZE,
    ):
        if file_size is None:
            file_size = os.path.getsize(file_path)

        super().__init__(
            file=file_path, file_size=file_size, max_file_part_size=max_file_part_size
        )


class FileLikePartsStrategy(BaseFilePartsStrategy):
    def __init__(
        self,
        file_name: str,
        file_like: IO,
        file_size: int,
        max_file_part_size: int = MIN_FILE_PART_SIZE,
    ):
        super().__init__(
            file=file_name, file_size=file_size, max_file_part_size=max_file_part_size
        )
        self.file_like = file_like



from .distributed_headers_context_manager import distributed_headers
from .span_context_manager import start_as_current_span
from .trace_context_manager import start_as_current_trace

__all__ = [
    "distributed_headers",
    "start_as_current_span",
    "start_as_current_trace",
]



"""Utility functions for usage report module."""

import datetime
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple


def aggregate_by_unit(metrics_response: Any, unit: str = "month") -> Dict[str, float]:
    """
    Aggregate metrics by specified time unit.

    Args:
        metrics_response: ProjectMetricResponsePublic object from get_project_metrics
        unit: Time unit for aggregation - "month", "week", "day", or "hour"

    Returns:
        Dictionary mapping time period key to total value
    """
    unit_data: Dict[str, float] = defaultdict(float)

    if metrics_response.results:
        for result in metrics_response.results:
            if result.data:
                for data_point in result.data:
                    if data_point.value is not None:
                        # Generate key based on unit
                        if unit == "month":
                            key = data_point.time.strftime("%Y-%m")
                        elif unit == "week":
                            # ISO week format: YYYY-Www
                            year, week, _ = data_point.time.isocalendar()
                            key = f"{year}-W{week:02d}"
                        elif unit == "day":
                            key = data_point.time.strftime("%Y-%m-%d")
                        elif unit == "hour":
                            key = data_point.time.strftime("%Y-%m-%d-%H")
                        else:
                            raise ValueError(f"Unsupported unit: {unit}")
                        unit_data[key] += data_point.value

    return dict(unit_data)


def format_datetime_key(dt: datetime.datetime, unit: str) -> str:
    """
    Format a datetime object to a key string based on the specified unit.

    Args:
        dt: Datetime object to format
        unit: Time unit - "month", "week", "day", or "hour"

    Returns:
        Formatted key string
    """
    if unit == "month":
        return dt.strftime("%Y-%m")
    elif unit == "week":
        year, week, _ = dt.isocalendar()
        return f"{year}-W{week:02d}"
    elif unit == "day":
        return dt.strftime("%Y-%m-%d")
    elif unit == "hour":
        return dt.strftime("%Y-%m-%d-%H")
    else:
        raise ValueError(f"Unsupported unit: {unit}")


def parse_and_normalize_datetime(
    dt_str: Any, reference_tz: Optional[datetime.tzinfo]
) -> Optional[datetime.datetime]:
    """
    Parse a datetime string and normalize it with respect to a reference timezone.

    Args:
        dt_str: Datetime string or datetime object to parse
        reference_tz: Reference timezone to use for naive datetimes

    Returns:
        Parsed datetime object, or None if parsing fails
    """
    if not dt_str:
        return None

    try:
        # If already a datetime object, return it
        if isinstance(dt_str, datetime.datetime):
            return dt_str

        # Parse ISO format datetime string
        if isinstance(dt_str, str):
            # Handle with or without timezone
            if "T" in dt_str:
                if dt_str.endswith("Z"):
                    dt = datetime.datetime.fromisoformat(dt_str.replace("Z", "+00:00"))
                elif "+" in dt_str or dt_str.count("-") > 2:
                    dt = datetime.datetime.fromisoformat(dt_str)
                else:
                    # Naive datetime
                    dt = datetime.datetime.fromisoformat(dt_str)
                    if reference_tz is not None:
                        # Make naive date timezone-aware
                        dt = dt.replace(tzinfo=reference_tz)
                return dt
            else:
                # Not a valid datetime string
                return None
        else:
            return dt_str
    except (ValueError, TypeError):
        return None


def normalize_timezone_for_comparison(
    dt: datetime.datetime,
    query_start_date: datetime.datetime,
    query_end_date: datetime.datetime,
) -> Tuple[datetime.datetime, datetime.datetime, datetime.datetime]:
    """
    Normalize timezones for date comparison.

    Args:
        dt: Datetime to normalize
        query_start_date: Start date for comparison
        query_end_date: End date for comparison

    Returns:
        Tuple of (normalized_dt, normalized_start_date, normalized_end_date)
    """
    # Handle timezone differences
    if dt.tzinfo is None and query_start_date.tzinfo is not None:
        dt = dt.replace(tzinfo=query_start_date.tzinfo)
        start_date_aware = query_start_date
        end_date_aware = query_end_date
    elif dt.tzinfo is not None and query_start_date.tzinfo is None:
        start_date_aware = query_start_date.replace(tzinfo=dt.tzinfo)
        end_date_aware = query_end_date.replace(tzinfo=dt.tzinfo)
    else:
        start_date_aware = query_start_date
        end_date_aware = query_end_date

    return dt, start_date_aware, end_date_aware


def extract_metric_data(
    projects: List[Dict[str, Any]],
    all_periods: List[str],
    metric_key: str,
    aggregation_fn: Optional[Callable[[Any], float]] = None,
) -> List[List[float]]:
    """
    Extract metric data from projects for all periods.

    Args:
        projects: List of project dictionaries with metrics_by_unit
        all_periods: List of period keys (e.g., "2024-01", "2024-02")
        metric_key: Key to extract from period_metrics (e.g., "trace_count", "cost")
        aggregation_fn: Optional function to aggregate metric values.
                       If None, uses default: sum dict values or use scalar value.

    Returns:
        List of lists, where each inner list contains metric values for one period
        across all projects
    """
    metric_data = []
    for period in all_periods:
        period_values = []
        for project in projects:
            period_metrics = project["metrics_by_unit"].get(period, {})
            metric_value = period_metrics.get(metric_key, 0)

            if aggregation_fn:
                metric_value = aggregation_fn(metric_value)
            elif isinstance(metric_value, dict):
                # Default: sum all dict values
                metric_value = sum(metric_value.values()) if metric_value else 0

            period_values.append(float(metric_value) if metric_value else 0.0)
        metric_data.append(period_values)

    return metric_data


def process_experiment_for_stats(
    experiment_dict: Dict[str, Any],
    experiment_by_unit: Dict[str, int],
    all_dates: List[datetime.datetime],
    query_start_date: datetime.datetime,
    query_end_date: datetime.datetime,
    unit: str,
    start_date: Optional[datetime.datetime],
) -> Tuple[int, int, int]:
    """
    Process a single experiment dictionary and update statistics.

    Returns:
        Tuple of (in_range_count, without_date_count, outside_range_count)
    """
    in_range = 0
    without_date = 0
    outside_range = 0

    # Extract created_at from raw dict (handles missing fields gracefully)
    created_at_str = experiment_dict.get("created_at")
    if created_at_str:
        # Parse datetime using helper function
        reference_tz = start_date.tzinfo if start_date else None
        exp_date = parse_and_normalize_datetime(created_at_str, reference_tz)

        if exp_date is None:
            without_date = 1
        else:
            # Normalize timezones for comparison
            exp_date, start_date_aware, end_date_aware = (
                normalize_timezone_for_comparison(
                    exp_date, query_start_date, query_end_date
                )
            )

            # Check if within date range
            if exp_date.tzinfo is not None:
                date_check = start_date_aware <= exp_date <= end_date_aware
            else:
                date_check = query_start_date <= exp_date <= query_end_date

            if date_check:
                in_range = 1
                all_dates.append(exp_date)
                unit_key = format_datetime_key(exp_date, unit)
                experiment_by_unit[unit_key] += 1
            else:
                outside_range = 1
    else:
        without_date = 1

    return (in_range, without_date, outside_range)



import logging
from typing import TYPE_CHECKING, Any, Dict, Optional

import opik
from opik import _logging as opik_logging
from opik import llm_usage, logging_messages
from . import provider_usage_extractor_protocol, langchain_run_helpers
from .langchain_run_helpers import langchain_usage

if TYPE_CHECKING:
    pass


LOGGER = logging.getLogger(__name__)


GROQ_CANDIDATE_USAGE_KEYS = {"prompt_tokens", "completion_tokens", "total_tokens"}


class GroqUsageExtractor(
    provider_usage_extractor_protocol.ProviderUsageExtractorProtocol
):
    PROVIDER = opik.LLMProvider.GROQ

    def is_provider_run(self, run_dict: Dict[str, Any]) -> bool:
        try:
            if run_dict.get("serialized") is None:
                return False

            serialized_kwargs = run_dict["serialized"].get("kwargs", {})
            has_groq_key = "groq_api_key" in serialized_kwargs

            return has_groq_key

        except Exception:
            LOGGER.debug(
                "Failed to check if Run instance is from Groq LLM, returning False.",
                exc_info=True,
            )
            return False

    def get_llm_usage_info(self, run_dict: Dict[str, Any]) -> llm_usage.LLMUsageInfo:
        opik_usage = _try_get_token_usage(run_dict)
        model = _try_get_model_name(run_dict)

        return llm_usage.LLMUsageInfo(
            provider=self.PROVIDER, model=model, usage=opik_usage
        )


def _try_get_token_usage(run_dict: Dict[str, Any]) -> Optional[llm_usage.OpikUsage]:
    """
    Attempts to extract and return the token usage from the given run dictionary.

    Depending on the execution type (invoke, streaming mode, async, etc.), or even the model name itself,
    token usage info might be in different places, different formats, or completely missing.
    """
    try:
        if token_usage := langchain_run_helpers.try_to_get_usage_by_search(
            run_dict, GROQ_CANDIDATE_USAGE_KEYS
        ):
            if isinstance(token_usage, dict):
                return llm_usage.OpikUsage.from_groq_completions_dict(token_usage)
            elif isinstance(token_usage, langchain_usage.LangChainUsage):
                # streaming mode handling
                # token usage data MAY be available at the end of streaming
                # in async mode may not provide token usage info
                groq_usage_dict = token_usage.map_to_groq_completions_usage()
                return llm_usage.OpikUsage.from_groq_completions_dict(groq_usage_dict)

        opik_logging.log_once_at_level(
            logging.WARNING,
            "Failed to extract token usage from presumably Groq LLM langchain run. Run dict: %s",
            LOGGER,
            run_dict,
        )

    except Exception:
        LOGGER.warning(
            logging_messages.FAILED_TO_EXTRACT_TOKEN_USAGE_FROM_PRESUMABLY_LANGCHAIN_GROQ_LLM_RUN,
            run_dict,
            exc_info=True,
        )
    return None


def _try_get_model_name(run_dict: Dict[str, Any]) -> Optional[str]:
    model = run_dict["extra"].get("metadata", {}).get("ls_model_name")
    if model is not None:
        model = model.split("/")[-1]

    return model



"""Healthcheck command for Opik CLI."""

from typing import Optional

import click

import opik.healthcheck as opik_healthcheck
from opik.cli.healthcheck.smoke_test import run_smoke_test


@click.command(context_settings={"ignore_unknown_options": True})
@click.option(
    "--show-installed-packages",
    is_flag=True,
    default=False,
    help="Print the list of installed packages to the console.",
)
@click.option(
    "--smoke-test",
    type=str,
    default=None,
    help="Run a smoke test to verify Opik integration is working correctly. Requires WORKSPACE value.",
    metavar="WORKSPACE",
)
@click.option(
    "--project-name",
    type=str,
    default=None,
    help="Project name for the smoke test (only used when --smoke-test is provided). Defaults to 'smoke-test-project'.",
)
@click.pass_context
def healthcheck(
    ctx: click.Context,
    show_installed_packages: bool,
    smoke_test: Optional[str],
    project_name: Optional[str],
) -> None:
    """
    Performs a health check of the application, including validation of configuration,
    verification of library installations, and checking the availability of the backend workspace.
    Prints all relevant information to assist in debugging and diagnostics.

    When --smoke-test is used, runs a smoke test to verify Opik integration is working correctly.
    The smoke test performs a basic sanity test by:
    - Creating a trace with random data
    - Adding a tracked LLM completion span
    - Attaching a dynamically generated test image
    - Verifying data is sent to Opik

    Examples:

      Run standard healthcheck:

        opik healthcheck

      Run healthcheck with smoke test:

        opik healthcheck --smoke-test my-workspace

      Run smoke test with custom project name:

        opik healthcheck --smoke-test my-workspace --project-name my-test-project
    """
    # Validate that --project-name is only used with --smoke-test
    if project_name is not None and smoke_test is None:
        raise click.BadParameter(
            "--project-name can only be used with --smoke-test",
            param_hint="--project-name",
        )

    if smoke_test is not None:
        # Use default project name if smoke_test is provided but project_name is not
        if project_name is None:
            project_name = "smoke-test-project"

        # Validate that workspace is not empty
        if smoke_test == "":
            raise click.BadParameter(
                "--smoke-test requires a non-empty workspace name",
                param_hint="--smoke-test",
            )

        # Get API key and debug flag from context
        api_key = ctx.obj.get("api_key") if ctx.obj else None
        debug = ctx.obj.get("debug", False) if ctx.obj else False

        try:
            run_smoke_test(
                workspace=smoke_test,
                project_name=project_name,
                api_key=api_key,
                debug=debug,
            )
        except Exception as e:
            click.echo(f"Smoke test failed: {e}", err=True)
            if debug:
                import traceback

                click.echo("Traceback:", err=True)
                click.echo(traceback.format_exc(), err=True)

    # Always run the standard healthcheck
    opik_healthcheck.run(show_installed_packages)



from typing import List

import pydantic

from ..metrics import score_result


class ThreadEvaluationResult(pydantic.BaseModel):
    """Evaluation result for a particular thread."""

    thread_id: str
    scores: List[score_result.ScoreResult] = pydantic.Field(default_factory=list)


class ThreadsEvaluationResult(pydantic.BaseModel):
    """Threads evaluation results"""

    results: List[ThreadEvaluationResult] = pydantic.Field(default_factory=list)



import logging
from typing import Optional, Literal

from opik.rest_api import client as rest_api_client
from opik.rest_api import types as rest_api_types

LOGGER = logging.getLogger(__name__)


class Optimization:
    def __init__(
        self,
        id: str,
        rest_client: rest_api_client.OpikApi,
    ) -> None:
        self._id = id
        self._rest_client = rest_client

    @property
    def id(self) -> str:
        return self._id

    def update(
        self,
        name: Optional[str] = None,
        status: Optional[
            Literal["running", "completed", "cancelled", "initialized", "error"]
        ] = None,
    ) -> None:
        LOGGER.debug(
            f"Updating optimization {self.id} with name {name} and status {status}"
        )
        self._rest_client.optimizations.update_optimizations_by_id(
            id=self.id,
            name=name,
            status=status,
        )

    def fetch_content(self) -> rest_api_types.OptimizationPublic:
        LOGGER.debug(f"Fetching optimization data {self.id}")
        return self._rest_client.optimizations.get_optimization_by_id(id=self.id)



"""Multi-turn simulation functionality."""

from typing import Callable, Optional, Dict, Any, List
from opik import id_helpers, track
from .simulated_user import SimulatedUser


def run_simulation(
    app: Callable,
    user_simulator: SimulatedUser,
    initial_message: Optional[str] = None,
    max_turns: int = 5,
    thread_id: Optional[str] = None,
    project_name: Optional[str] = None,
    **app_kwargs: Any,
) -> Dict[str, Any]:
    """
    Run a multi-turn conversation simulation between a simulated user and an app.

    1. The simulator passes single message strings to the app
    2. The app manages full conversation history internally using thread_id
    3. The app logs traces with thread_id for evaluation

    Args:
        app: Callable that processes messages and manages conversation history internally.
            Signature: app(message: str, *, thread_id: str, **kwargs) -> Dict[str, str]
            The app is automatically decorated with @track and thread_id is injected via opik_args.
        user_simulator: SimulatedUser instance that generates user responses
        initial_message: Optional initial message from the user. If None, generated by simulator
        max_turns: Maximum number of conversation turns (default: 5)
        thread_id: Optional thread ID for grouping traces. Generated if not provided
        project_name: Optional project name for trace logging
        **app_kwargs: Additional keyword arguments passed to the app

    Returns:
        Dict containing:
        - thread_id: The thread ID used for this simulation
        - conversation_history: List of message dicts from the simulation
        - project_name: Project name if provided
    """
    # Generate thread_id if not provided
    if thread_id is None:
        thread_id = id_helpers.generate_id()

    # Automatically decorate app if not already decorated
    if not hasattr(app, "opik_tracked"):
        app_name = app.__name__ if hasattr(app, "__name__") else "simulation_app"
        app = track(name=app_name)(app)

    # Track conversation for simulator (app manages its own history internally)
    conversation_history: List[Dict[str, str]] = []

    # Generate initial message if needed
    if initial_message is None:
        initial_message = user_simulator.generate_response(conversation_history)

    # Simulation loop
    for turn in range(max_turns):
        # Get user message
        if turn == 0:
            user_message_text = initial_message
        else:
            user_message_text = user_simulator.generate_response(conversation_history)

        # Create message dict for tracking
        user_message = {"role": "user", "content": user_message_text}
        conversation_history.append(user_message)

        # Call app with SINGLE message string, thread_id parameter, and opik_args for tracing
        try:
            assistant_message = app(
                user_message_text,
                thread_id=thread_id,
                **app_kwargs,
                opik_args={
                    "trace": {
                        "thread_id": thread_id,
                        "metadata": {"turn": turn + 1, "project_name": project_name},
                    }
                },
            )
        except Exception as e:
            # Handle app errors gracefully
            assistant_message = {
                "role": "assistant",
                "content": f"Error processing message: {str(e)}",
            }

        # Validate assistant message format
        if (
            not isinstance(assistant_message, dict)
            or "role" not in assistant_message
            or "content" not in assistant_message
        ):
            assistant_message = {
                "role": "assistant",
                "content": str(assistant_message)
                if assistant_message
                else "No response",
            }

        conversation_history.append(assistant_message)

    return {
        "thread_id": thread_id,
        "conversation_history": conversation_history,
        "project_name": project_name,
    }



"""
Opik integration for Harbor benchmark evaluation framework.

Example:
    >>> from opik.integrations.harbor import track_harbor
    >>> job = Job(config)
    >>> tracked_job = track_harbor(job)
    >>> result = await tracked_job.run()

Or enable tracking globally (for CLI usage):
    >>> from opik.integrations.harbor import track_harbor
    >>> track_harbor()
"""

from .opik_tracker import track_harbor, reset_harbor_tracking

__all__ = ["track_harbor", "reset_harbor_tracking"]



import logging
from typing import Any, Dict, List

LOGGER = logging.getLogger(__name__)


def _handle_message_start(event: Dict[str, Any], result: Dict[str, Any]) -> None:
    """Extract role from messageStart event."""
    message_start = event.get("messageStart")
    if isinstance(message_start, dict):
        role = message_start.get("role")
        if role:
            result["output"]["message"]["role"] = role


def _handle_content_block_delta(event: Dict[str, Any], result: Dict[str, Any]) -> None:
    """
    Extract content from contentBlockDelta event.

    Handles multiple delta types:
    - delta.text: Regular text streaming
    - delta.toolUse: Structured output / tool calls (Issue #3829)
    """
    content_block_delta = event.get("contentBlockDelta")
    if not isinstance(content_block_delta, dict):
        return

    delta = content_block_delta.get("delta")
    if not isinstance(delta, dict):
        return

    content = result["output"]["message"]["content"][0]

    # Handle regular text streaming
    if "text" in delta:
        content["text"] += delta["text"]
        return

    # Handle structured output / tool use (Issue #3829)
    # Ref: https://github.com/comet-ml/opik/issues/3829
    if "toolUse" in delta:
        if "toolUse" not in content:
            content["toolUse"] = {}
        content["toolUse"].update(delta["toolUse"])
        return

    # Log other delta types for future compatibility
    LOGGER.debug("Unknown delta type in contentBlockDelta: %s", list(delta.keys()))


def _handle_message_stop(event: Dict[str, Any], result: Dict[str, Any]) -> None:
    """Extract stopReason from messageStop event."""
    message_stop = event.get("messageStop")
    if isinstance(message_stop, dict):
        stop_reason = message_stop.get("stopReason")
        if stop_reason:
            result["stopReason"] = stop_reason


def _handle_metadata(event: Dict[str, Any], result: Dict[str, Any]) -> None:
    """Extract usage and metrics from metadata event."""
    metadata = event.get("metadata")
    if not isinstance(metadata, dict):
        return

    # Extract usage information
    if "usage" in metadata:
        result["usage"] = metadata["usage"]

    # Extract metrics information
    if "metrics" in metadata:
        metrics = metadata["metrics"]
        if isinstance(metrics, dict) and "latencyMs" in metrics:
            result["metrics"] = {"latencyMs": metrics["latencyMs"]}


def aggregate_converse_stream_chunks(items: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    Aggregate streaming chunks from AWS Bedrock converse_stream API into a single response.

    This function handles various event structures from different Bedrock model providers:
    - Anthropic (Claude): Standard messageStart, contentBlockDelta, messageStop events
    - Amazon (Nova): Amazon's proprietary event format with potential variations
    - Meta (Llama): Open-source model events with different tokenization patterns
    - Mistral (Pixtral): Multimodal model events that may include additional content types
    - DeepSeek (R1): Reasoning model events with extended thought processes (OPIK-2910 fix)

    Event Structure Variations by Provider:
    ========================================

    Standard Converse Stream Events (from AWS documentation):
    - messageStart: Contains role and initial metadata
    - contentBlockStart: Marks beginning of content block
    - contentBlockDelta: Contains incremental text in delta.text
    - contentBlockStop: Marks end of content block
    - messageStop: Contains stopReason (e.g., "end_turn", "stop_sequence")
    - metadata: Contains usage stats and performance metrics

    Known Variations:
    - DeepSeek R1: May have different delta structure or additional reasoning fields (OPIK-2910)
    - Tool Use/Structured Output: delta.toolUse instead of delta.text (Issue #3829)
    - Multimodal models: May include non-text content blocks (images, documents)
    - Different models: Varying field names, nesting levels, or optional fields

    References:
    - AWS Bedrock Converse API: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html
    - Streaming Events: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ConverseStreamOutput.html
    - ContentBlockDelta: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ContentBlockDelta.html
    - Tool Use Guide: https://docs.aws.amazon.com/bedrock/latest/userguide/tool-use.html
    - DeepSeek R1 Issue: https://comet-ml.atlassian.net/browse/OPIK-2910
    - Tool Use Issue: https://github.com/comet-ml/opik/issues/3829

    Args:
        items: List of streaming event dictionaries from Bedrock converse_stream

    Returns:
        Aggregated response dictionary with structure:
        {
            "output": {
                "message": {
                    "role": "assistant",
                    "content": [{"text": "aggregated text"}]
                }
            },
            "stopReason": "end_turn",  # Optional
            "usage": {...},  # Optional
            "metrics": {"latencyMs": ...}  # Optional
        }
    """

    result: Dict[str, Any] = {
        "output": {"message": {"role": "assistant", "content": [{"text": ""}]}}
    }

    for event in items:
        if not isinstance(event, dict):
            LOGGER.debug("Skipping non-dictionary event: %s", type(event))
            continue

        try:
            if "messageStart" in event:
                _handle_message_start(event, result)

            if "contentBlockDelta" in event:
                _handle_content_block_delta(event, result)

            if "messageStop" in event:
                _handle_message_stop(event, result)

            if "metadata" in event:
                _handle_metadata(event, result)

        except Exception as e:
            LOGGER.warning(
                "Unexpected error processing event: %s. Event: %s",
                str(e),
                event,
                exc_info=True,
            )

    return result


def aggregate_invoke_agent_chunks(items: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    Aggregate streaming chunks from AWS Bedrock invoke_agent API.

    Note: The implementation uses a simplified approach as the completion payload
    only contains chunks without additional metadata (as of implementation date).

    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/bedrock-agent-runtime/client/invoke_agent.html

    Args:
        items: List of chunk event dictionaries from invoke_agent stream

    Returns:
        Aggregated response dictionary with decoded text output
    """
    merged_chunks = b""

    for item in items:
        if isinstance(item, dict) and "chunk" in item:
            chunk = item["chunk"]
            if isinstance(chunk, dict) and "bytes" in chunk:
                merged_chunks += chunk["bytes"]

    return {"output": merged_chunks.decode("utf-8")}



import dataclasses
import math
import statistics
from collections import defaultdict
from typing import List, Optional, Dict

from opik.evaluation import test_result


@dataclasses.dataclass
class ScoreStatistics:
    """Statistics for a single score metric across multiple trials."""

    mean: float
    max: float
    min: float
    values: List[float]
    std: Optional[float] = None  # Standard deviation (None if count < 2)


def calculate_aggregated_statistics(
    evaluation_results: List[test_result.TestResult],
) -> Dict[str, ScoreStatistics]:
    """
    Calculate mean, max, and min scores for each score name in the evaluation test results.

    Args:
        evaluation_results: List of TestResult objects to be aggregated

    Returns:
        Dict mapping score names to their aggregated statistics
    """
    if not evaluation_results:
        return {}

    # Group scores by name across all trials
    scores_by_name = defaultdict(list)

    for test_result_ in evaluation_results:
        for score_result in test_result_.score_results:
            # Only include successful scores with valid values
            if not score_result.scoring_failed and _is_valid_score_value(
                score_result.value
            ):
                scores_by_name[score_result.name].append(score_result.value)

    # Calculate aggregated statistics for each score name
    aggregated_scores = {}
    for score_name, values in scores_by_name.items():
        if values:
            std = statistics.stdev(values) if len(values) >= 2 else None

            aggregated_scores[score_name] = ScoreStatistics(
                mean=statistics.mean(values),
                max=max(values),
                min=min(values),
                values=values.copy(),  # Store the actual values used
                std=std,
            )

    return aggregated_scores


def _is_valid_score_value(value: float) -> bool:
    """Check if a score value is valid for statistical calculations."""
    return isinstance(value, (int, float)) and math.isfinite(value)



import abc
from typing import Optional

from opik.message_processing import messages


class MessagePreprocessor(abc.ABC):
    """
    Abstract base class for message preprocessing.

    This class provides a common interface for pre-processing messages, allowing
    derived classes to implement custom preprocessing logic tailored to specific
    requirements. Instances of this class cannot be created directly; it must be
    subclassed with the `preprocess` method implemented.
    """

    @abc.abstractmethod
    def preprocess(
        self, message: Optional[messages.BaseMessage]
    ) -> Optional[messages.BaseMessage]:
        """
        Processes and preprocesses the given message to prepare it for further operations.

        This is an abstract method and needs to be implemented in any concrete subclass. The
        preprocessing step is typically used for transformations or checks on the given input
        message before further processing.

        Args:
            message: The input message to be preprocessed. This can
                optionally be None.

        Returns:
            The processed message after preprocessing. Returns None if the input message is None
            or if a message was fully consumed here and no further processing is required.
        """
        pass



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.attachment_page import AttachmentPage
from ..types.complete_multipart_upload_request_entity_type import CompleteMultipartUploadRequestEntityType
from ..types.multipart_upload_part import MultipartUploadPart
from ..types.start_multipart_upload_response import StartMultipartUploadResponse
from .raw_client import AsyncRawAttachmentsClient, RawAttachmentsClient
from .types.attachment_list_request_entity_type import AttachmentListRequestEntityType
from .types.download_attachment_request_entity_type import DownloadAttachmentRequestEntityType
from .types.start_multipart_upload_request_entity_type import StartMultipartUploadRequestEntityType
from .types.upload_attachment_request_entity_type import UploadAttachmentRequestEntityType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class AttachmentsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawAttachmentsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawAttachmentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawAttachmentsClient
        """
        return self._raw_client

    def attachment_list(
        self,
        *,
        project_id: str,
        entity_type: AttachmentListRequestEntityType,
        entity_id: str,
        path: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AttachmentPage:
        """
        Attachments list for entity

        Parameters
        ----------
        project_id : str

        entity_type : AttachmentListRequestEntityType

        entity_id : str

        path : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AttachmentPage
            Attachment Resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.attachments.attachment_list(project_id='project_id', entity_type="trace", entity_id='entity_id', path='path', )
        """
        _response = self._raw_client.attachment_list(
            project_id=project_id,
            entity_type=entity_type,
            entity_id=entity_id,
            path=path,
            page=page,
            size=size,
            request_options=request_options,
        )
        return _response.data

    def complete_multi_part_upload(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Complete multipart attachment upload

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import MultipartUploadPart
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.attachments.complete_multi_part_upload(file_name='file_name', entity_type="trace", entity_id='entity_id', file_size=1000000, upload_id='upload_id', uploaded_file_parts=[MultipartUploadPart(e_tag='e_tag', part_number=1, )], )
        """
        _response = self._raw_client.complete_multi_part_upload(
            file_name=file_name,
            entity_type=entity_type,
            entity_id=entity_id,
            file_size=file_size,
            upload_id=upload_id,
            uploaded_file_parts=uploaded_file_parts,
            project_name=project_name,
            container_id=container_id,
            mime_type=mime_type,
            request_options=request_options,
        )
        return _response.data

    def delete_attachments(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Delete attachments

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import MultipartUploadPart
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.attachments.delete_attachments(file_name='file_name', entity_type="trace", entity_id='entity_id', file_size=1000000, upload_id='upload_id', uploaded_file_parts=[MultipartUploadPart(e_tag='e_tag', part_number=1, )], )
        """
        _response = self._raw_client.delete_attachments(
            file_name=file_name,
            entity_type=entity_type,
            entity_id=entity_id,
            file_size=file_size,
            upload_id=upload_id,
            uploaded_file_parts=uploaded_file_parts,
            project_name=project_name,
            container_id=container_id,
            mime_type=mime_type,
            request_options=request_options,
        )
        return _response.data

    def download_attachment(
        self,
        *,
        container_id: str,
        entity_type: DownloadAttachmentRequestEntityType,
        entity_id: str,
        file_name: str,
        mime_type: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[bytes]:
        """
        Download attachment from MinIO

        Parameters
        ----------
        container_id : str

        entity_type : DownloadAttachmentRequestEntityType

        entity_id : str

        file_name : str

        mime_type : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[bytes]
            Attachment Resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.attachments.download_attachment(container_id='container_id', entity_type="trace", entity_id='entity_id', file_name='x', mime_type='x', )
        """
        with self._raw_client.download_attachment(
            container_id=container_id,
            entity_type=entity_type,
            entity_id=entity_id,
            file_name=file_name,
            mime_type=mime_type,
            workspace_name=workspace_name,
            request_options=request_options,
        ) as r:
            yield from r.data

    def start_multi_part_upload(
        self,
        *,
        file_name: str,
        num_of_file_parts: int,
        entity_type: StartMultipartUploadRequestEntityType,
        entity_id: str,
        path: str,
        mime_type: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> StartMultipartUploadResponse:
        """
        Start multipart attachment upload

        Parameters
        ----------
        file_name : str

        num_of_file_parts : int

        entity_type : StartMultipartUploadRequestEntityType

        entity_id : str

        path : str

        mime_type : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        StartMultipartUploadResponse
            MultipartUploadResponse

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.attachments.start_multi_part_upload(file_name='file_name', num_of_file_parts=1, entity_type="trace", entity_id='entity_id', path='path', )
        """
        _response = self._raw_client.start_multi_part_upload(
            file_name=file_name,
            num_of_file_parts=num_of_file_parts,
            entity_type=entity_type,
            entity_id=entity_id,
            path=path,
            mime_type=mime_type,
            project_name=project_name,
            request_options=request_options,
        )
        return _response.data

    def upload_attachment(
        self,
        *,
        file_name: str,
        entity_type: UploadAttachmentRequestEntityType,
        entity_id: str,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        project_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Upload attachment to MinIO

        Parameters
        ----------
        file_name : str

        entity_type : UploadAttachmentRequestEntityType

        entity_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        project_name : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.attachments.upload_attachment(file_name='file_name', entity_type="trace", entity_id='entity_id', request={'key': 'value'
        }, )
        """
        _response = self._raw_client.upload_attachment(
            file_name=file_name,
            entity_type=entity_type,
            entity_id=entity_id,
            request=request,
            project_name=project_name,
            mime_type=mime_type,
            request_options=request_options,
        )
        return _response.data


class AsyncAttachmentsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawAttachmentsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawAttachmentsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawAttachmentsClient
        """
        return self._raw_client

    async def attachment_list(
        self,
        *,
        project_id: str,
        entity_type: AttachmentListRequestEntityType,
        entity_id: str,
        path: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AttachmentPage:
        """
        Attachments list for entity

        Parameters
        ----------
        project_id : str

        entity_type : AttachmentListRequestEntityType

        entity_id : str

        path : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AttachmentPage
            Attachment Resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.attachments.attachment_list(project_id='project_id', entity_type="trace", entity_id='entity_id', path='path', )
        asyncio.run(main())
        """
        _response = await self._raw_client.attachment_list(
            project_id=project_id,
            entity_type=entity_type,
            entity_id=entity_id,
            path=path,
            page=page,
            size=size,
            request_options=request_options,
        )
        return _response.data

    async def complete_multi_part_upload(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Complete multipart attachment upload

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import MultipartUploadPart
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.attachments.complete_multi_part_upload(file_name='file_name', entity_type="trace", entity_id='entity_id', file_size=1000000, upload_id='upload_id', uploaded_file_parts=[MultipartUploadPart(e_tag='e_tag', part_number=1, )], )
        asyncio.run(main())
        """
        _response = await self._raw_client.complete_multi_part_upload(
            file_name=file_name,
            entity_type=entity_type,
            entity_id=entity_id,
            file_size=file_size,
            upload_id=upload_id,
            uploaded_file_parts=uploaded_file_parts,
            project_name=project_name,
            container_id=container_id,
            mime_type=mime_type,
            request_options=request_options,
        )
        return _response.data

    async def delete_attachments(
        self,
        *,
        file_name: str,
        entity_type: CompleteMultipartUploadRequestEntityType,
        entity_id: str,
        file_size: int,
        upload_id: str,
        uploaded_file_parts: typing.Sequence[MultipartUploadPart],
        project_name: typing.Optional[str] = OMIT,
        container_id: typing.Optional[str] = OMIT,
        mime_type: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Delete attachments

        Parameters
        ----------
        file_name : str

        entity_type : CompleteMultipartUploadRequestEntityType

        entity_id : str

        file_size : int

        upload_id : str

        uploaded_file_parts : typing.Sequence[MultipartUploadPart]

        project_name : typing.Optional[str]
            If null, the default project is used

        container_id : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import MultipartUploadPart
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.attachments.delete_attachments(file_name='file_name', entity_type="trace", entity_id='entity_id', file_size=1000000, upload_id='upload_id', uploaded_file_parts=[MultipartUploadPart(e_tag='e_tag', part_number=1, )], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_attachments(
            file_name=file_name,
            entity_type=entity_type,
            entity_id=entity_id,
            file_size=file_size,
            upload_id=upload_id,
            uploaded_file_parts=uploaded_file_parts,
            project_name=project_name,
            container_id=container_id,
            mime_type=mime_type,
            request_options=request_options,
        )
        return _response.data

    async def download_attachment(
        self,
        *,
        container_id: str,
        entity_type: DownloadAttachmentRequestEntityType,
        entity_id: str,
        file_name: str,
        mime_type: str,
        workspace_name: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[bytes]:
        """
        Download attachment from MinIO

        Parameters
        ----------
        container_id : str

        entity_type : DownloadAttachmentRequestEntityType

        entity_id : str

        file_name : str

        mime_type : str

        workspace_name : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[bytes]
            Attachment Resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.attachments.download_attachment(container_id='container_id', entity_type="trace", entity_id='entity_id', file_name='x', mime_type='x', )
        asyncio.run(main())
        """
        async with self._raw_client.download_attachment(
            container_id=container_id,
            entity_type=entity_type,
            entity_id=entity_id,
            file_name=file_name,
            mime_type=mime_type,
            workspace_name=workspace_name,
            request_options=request_options,
        ) as r:
            async for data in r.data:
                yield data

    async def start_multi_part_upload(
        self,
        *,
        file_name: str,
        num_of_file_parts: int,
        entity_type: StartMultipartUploadRequestEntityType,
        entity_id: str,
        path: str,
        mime_type: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> StartMultipartUploadResponse:
        """
        Start multipart attachment upload

        Parameters
        ----------
        file_name : str

        num_of_file_parts : int

        entity_type : StartMultipartUploadRequestEntityType

        entity_id : str

        path : str

        mime_type : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        StartMultipartUploadResponse
            MultipartUploadResponse

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.attachments.start_multi_part_upload(file_name='file_name', num_of_file_parts=1, entity_type="trace", entity_id='entity_id', path='path', )
        asyncio.run(main())
        """
        _response = await self._raw_client.start_multi_part_upload(
            file_name=file_name,
            num_of_file_parts=num_of_file_parts,
            entity_type=entity_type,
            entity_id=entity_id,
            path=path,
            mime_type=mime_type,
            project_name=project_name,
            request_options=request_options,
        )
        return _response.data

    async def upload_attachment(
        self,
        *,
        file_name: str,
        entity_type: UploadAttachmentRequestEntityType,
        entity_id: str,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        project_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Upload attachment to MinIO

        Parameters
        ----------
        file_name : str

        entity_type : UploadAttachmentRequestEntityType

        entity_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        project_name : typing.Optional[str]

        mime_type : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.attachments.upload_attachment(file_name='file_name', entity_type="trace", entity_id='entity_id', request={'key': 'value'
            }, )
        asyncio.run(main())
        """
        _response = await self._raw_client.upload_attachment(
            file_name=file_name,
            entity_type=entity_type,
            entity_id=entity_id,
            request=request,
            project_name=project_name,
            mime_type=mime_type,
            request_options=request_options,
        )
        return _response.data



"""Claude/Anthropic chunk aggregator."""

import json
import logging
from typing import Any, Dict, List

from .. import usage_converters
from .base import ChunkAggregator

LOGGER = logging.getLogger(__name__)


class ClaudeAggregator(ChunkAggregator):
    """
    Aggregator for Claude/Anthropic streaming format.

    Returns Claude's native message format with content blocks,
    with usage standardized to Bedrock format (camelCase).

    Claude chunk structure uses snake_case:
    - message_start: Contains role and initial usage
    - content_block_start: Content block metadata
    - content_block_delta: Text chunks in delta.text
    - content_block_stop: End of content block
    - message_delta: Stop reason and updated usage
    - message_stop: amazon-bedrock-invocationMetrics
    """

    def aggregate(self, items: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Aggregate Claude chunks into native Claude message structure with Bedrock usage."""
        LOGGER.debug("Claude aggregator processing %d items", len(items))

        content_blocks = []
        current_text = ""
        role = "assistant"
        stop_reason = None
        input_tokens = 0
        output_tokens = 0

        for item in items:
            if "chunk" not in item:
                continue

            try:
                chunk_data = json.loads(item["chunk"]["bytes"])
                chunk_type = chunk_data.get("type", "")

                if chunk_type == "message_start":
                    message = chunk_data.get("message", {})
                    role = message.get("role", "assistant")
                    usage = message.get("usage", {})
                    input_tokens = usage.get("input_tokens", 0)
                    output_tokens = usage.get("output_tokens", 0)
                    LOGGER.debug(
                        "Claude message_start: input_tokens=%d, output_tokens=%d",
                        input_tokens,
                        output_tokens,
                    )

                elif chunk_type == "content_block_delta":
                    delta = chunk_data.get("delta", {})
                    if "text" in delta:
                        current_text += delta["text"]

                elif chunk_type == "content_block_stop":
                    if current_text:
                        content_blocks.append({"type": "text", "text": current_text})
                        current_text = ""

                elif chunk_type == "message_delta":
                    delta = chunk_data.get("delta", {})
                    if "stop_reason" in delta:
                        stop_reason = delta["stop_reason"]
                    usage = chunk_data.get("usage", {})
                    if "output_tokens" in usage:
                        output_tokens = usage["output_tokens"]
                        LOGGER.debug(
                            "Claude message_delta: output_tokens=%d", output_tokens
                        )

                elif chunk_type == "message_stop":
                    metrics = chunk_data.get("amazon-bedrock-invocationMetrics", {})
                    if metrics:
                        input_tokens = metrics.get("inputTokenCount", input_tokens)
                        output_tokens = metrics.get("outputTokenCount", output_tokens)
                        LOGGER.debug(
                            "Claude bedrock metrics: input=%d, output=%d",
                            input_tokens,
                            output_tokens,
                        )

            except (json.JSONDecodeError, KeyError, TypeError) as e:
                LOGGER.debug("Claude aggregator error processing chunk: %s", e)
                continue

        # Add any remaining text
        if current_text:
            content_blocks.append({"type": "text", "text": current_text})

        total_text = "".join(
            block["text"] for block in content_blocks if block["type"] == "text"
        )
        LOGGER.debug(
            "Claude aggregated: %d chars, input_tokens=%d, output_tokens=%d",
            len(total_text),
            input_tokens,
            output_tokens,
        )

        # Convert to Bedrock usage format using shared converter
        bedrock_usage = usage_converters.anthropic_to_bedrock_usage(
            {"input_tokens": input_tokens, "output_tokens": output_tokens}
        )

        # Return Claude's native format with Bedrock usage
        return {
            "role": role,
            "content": content_blocks,
            "stop_reason": stop_reason,
            "usage": bedrock_usage,
        }



"""Common utilities for import functionality."""

from collections import deque
from typing import Any, Dict, List, Optional

import opik
from opik.types import FeedbackScoreDict
from rich.console import Console
from rich.table import Table

console = Console()


def debug_print(message: str, debug: bool) -> None:
    """Print debug message only if debug is enabled."""
    if debug:
        console.print(f"[blue]{message}[/blue]")


def matches_name_pattern(name: str, pattern: Optional[str]) -> bool:
    """Check if a name matches the given pattern using case-insensitive substring matching."""
    if pattern is None:
        return True
    # Simple string matching - check if pattern is contained in name (case-insensitive)
    return pattern.lower() in name.lower()


def translate_trace_id(
    original_trace_id: str, trace_id_map: Dict[str, str]
) -> Optional[str]:
    """Translate an original trace id from export to the newly created id.

    Args:
        original_trace_id: The original trace ID from the export
        trace_id_map: Mapping from original trace IDs to new trace IDs (required, can be empty)

    Returns:
        The new trace ID if found in the map, None otherwise.
    """
    return trace_id_map.get(original_trace_id)


def find_dataset_item_by_content(
    dataset: opik.Dataset, expected_content: Dict[str, Any]
) -> Optional[str]:
    """Find a dataset item by matching its content.

    Compares all fields in expected_content (excluding 'id') with dataset items.
    """
    try:
        items = dataset.get_items()
        # Remove 'id' from expected_content for comparison
        expected_without_id = {k: v for k, v in expected_content.items() if k != "id"}

        for item in items:
            # Compare all fields (excluding 'id')
            item_without_id = {k: v for k, v in item.items() if k != "id"}
            if item_without_id == expected_without_id:
                return item.get("id")
    except Exception:
        pass
    return None


def create_dataset_item(dataset: opik.Dataset, item_data: Dict[str, Any]) -> str:
    """Create a dataset item and return its ID.

    Uses all fields from item_data (except 'id') to support flexible dataset schemas.
    """
    # Use all fields from item_data except 'id' (which is handled separately)
    new_item = {k: v for k, v in item_data.items() if k != "id" and v is not None}

    # Ensure item is not empty (backend requires non-empty data)
    if not new_item:
        raise ValueError(
            "Dataset item data is empty - at least one field must be provided"
        )

    dataset.insert([new_item])

    # Find the newly created item by matching all fields
    items = dataset.get_items()
    for item in items:
        # Compare all fields (excluding 'id')
        item_without_id = {k: v for k, v in item.items() if k != "id"}
        if item_without_id == new_item:
            item_id = item.get("id")
            if item_id is not None:
                return item_id

    dataset_name = getattr(dataset, "name", None)
    dataset_info = f", Dataset: {dataset_name!r}" if dataset_name else ""
    raise Exception(
        f"Failed to create dataset item. Content: {new_item!r}{dataset_info}"
    )


def handle_trace_reference(item_data: Dict[str, Any]) -> Optional[str]:
    """Handle trace references from deduplicated exports."""
    trace_reference = item_data.get("trace_reference")
    if trace_reference:
        trace_id = trace_reference.get("trace_id")
        return trace_id

    # Fall back to direct trace_id
    return item_data.get("trace_id")


def clean_usage_for_import(
    usage: Optional[Dict[str, Any]],
) -> Optional[Dict[str, Any]]:
    """Return usage data as-is for import.

    When exporting, usage data is already in backend-compatible format with:
    - Top-level OpenAI-formatted keys: completion_tokens, prompt_tokens, total_tokens
    - Provider-specific keys with 'original_usage.' prefix

    The SDK's validation_helpers.validate_and_parse_usage now detects when usage
    data contains 'original_usage.' prefixed keys and passes it through unchanged,
    preserving the original values.

    Args:
        usage: The usage dictionary from exported data, or None

    Returns:
        The usage dictionary as-is (already in backend format), or None if input was None
    """
    # Return usage as-is - the SDK will detect the 'original_usage.' prefix
    # and pass it through without reprocessing
    return usage


def clean_feedback_scores(
    feedback_scores: Optional[List[Dict[str, Any]]],
) -> Optional[List[FeedbackScoreDict]]:
    """Clean feedback scores by removing fields that are not allowed when creating them.

    Exported feedback scores include read-only fields like 'source', 'created_at', etc.
    that must be removed before importing.

    Allowed fields: name, value, category_name, reason
    """
    if not feedback_scores:
        return None

    cleaned_scores: List[FeedbackScoreDict] = []
    for score in feedback_scores:
        if not isinstance(score, dict):
            continue

        # Only keep allowed fields
        name = score.get("name")
        value = score.get("value")

        # Only add if name and value are present
        if not name or value is None:
            continue

        # Construct FeedbackScoreDict with required fields
        cleaned_score: FeedbackScoreDict = {
            "name": name,
            "value": value,
        }

        # Add optional fields if they exist
        if "category_name" in score:
            cleaned_score["category_name"] = score.get("category_name")
        if "reason" in score:
            cleaned_score["reason"] = score.get("reason")

        cleaned_scores.append(cleaned_score)

    return cleaned_scores if cleaned_scores else None


def sort_spans_topologically(spans_info: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Sort spans in topological order to ensure parents are processed before children.

    This function performs a level-order traversal (breadth-first) of the span tree,
    ensuring that all parent spans are processed before their children, regardless
    of how deep the hierarchy goes.

    Args:
        spans_info: List of span dictionaries, each with 'id' and optional 'parent_span_id'

    Returns:
        List of spans sorted in topological order (parents before children)
    """
    if not spans_info:
        return []

    # Build children map: parent_id -> list of child spans
    # Also track all span IDs to detect missing parent references
    children_map: Dict[str, List[Dict[str, Any]]] = {}
    root_spans: List[Dict[str, Any]] = []
    all_span_ids = {span.get("id") for span in spans_info if span.get("id")}

    for span in spans_info:
        parent_span_id = span.get("parent_span_id")

        if not parent_span_id:
            # Root span (no parent)
            root_spans.append(span)
        elif parent_span_id in all_span_ids:
            # Child span with valid parent reference - add to children map
            if parent_span_id not in children_map:
                children_map[parent_span_id] = []
            children_map[parent_span_id].append(span)
        else:
            # Span references a parent that doesn't exist - treat as root span
            # This handles corrupted data gracefully
            root_spans.append(span)

    # Perform level-order traversal (breadth-first)
    result: List[Dict[str, Any]] = []
    visited: set[str] = set()

    # If no root spans exist (e.g., cycles), seed queue with all spans
    # This ensures we process all spans even in cyclic or disconnected graphs
    queue: deque[Dict[str, Any]]
    if not root_spans:
        queue = deque(spans_info)
    else:
        queue = deque(root_spans)

    while queue:
        current = queue.popleft()
        current_id = current.get("id")

        # Skip if already visited (prevents infinite loops in cycles)
        if current_id and current_id in visited:
            continue

        result.append(current)
        if current_id:
            visited.add(current_id)

        # Add all children of current span to queue (only if not visited)
        if current_id and current_id in children_map:
            for child in children_map[current_id]:
                child_id = child.get("id")
                if child_id and child_id not in visited:
                    queue.append(child)

    # Append any spans not yet visited (handles disconnected components)
    for span in spans_info:
        span_id = span.get("id")
        if span_id and span_id not in visited:
            result.append(span)

    return result


def print_import_summary(stats: Dict[str, int]) -> None:
    """Print a nice summary table of import statistics."""
    table = Table(
        title="📥 Import Summary", show_header=True, header_style="bold magenta"
    )
    table.add_column("Type", style="cyan", no_wrap=True)
    table.add_column("Imported", justify="right", style="green")
    table.add_column("Skipped", justify="right", style="yellow")
    table.add_column("Errors", justify="right", style="red")

    # Add rows for each type
    if (
        stats.get("experiments", 0) > 0
        or stats.get("experiments_skipped", 0) > 0
        or stats.get("experiments_errors", 0) > 0
    ):
        imported = stats.get("experiments", 0)
        skipped = stats.get("experiments_skipped", 0)
        errors = stats.get("experiments_errors", 0)
        table.add_row(
            "🧪 Experiments",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("datasets", 0) > 0
        or stats.get("datasets_skipped", 0) > 0
        or stats.get("datasets_errors", 0) > 0
    ):
        imported = stats.get("datasets", 0)
        skipped = stats.get("datasets_skipped", 0)
        errors = stats.get("datasets_errors", 0)
        table.add_row(
            "📊 Datasets",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("traces", 0) > 0
        or stats.get("traces_skipped", 0) > 0
        or stats.get("traces_errors", 0) > 0
    ):
        imported = stats.get("traces", 0)
        skipped = stats.get("traces_skipped", 0)
        errors = stats.get("traces_errors", 0)
        table.add_row(
            "🔍 Traces",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("prompts", 0) > 0
        or stats.get("prompts_skipped", 0) > 0
        or stats.get("prompts_errors", 0) > 0
    ):
        imported = stats.get("prompts", 0)
        skipped = stats.get("prompts_skipped", 0)
        errors = stats.get("prompts_errors", 0)
        table.add_row(
            "💬 Prompts",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    if (
        stats.get("projects", 0) > 0
        or stats.get("projects_skipped", 0) > 0
        or stats.get("projects_errors", 0) > 0
    ):
        imported = stats.get("projects", 0)
        skipped = stats.get("projects_skipped", 0)
        errors = stats.get("projects_errors", 0)
        table.add_row(
            "📁 Projects",
            str(imported),
            str(skipped) if skipped > 0 else "",
            str(errors) if errors > 0 else "",
        )

    # Calculate totals
    total_imported = sum(
        [
            stats.get(key, 0)
            for key in ["experiments", "datasets", "traces", "prompts", "projects"]
        ]
    )
    total_skipped = sum(
        [
            stats.get(key, 0)
            for key in [
                "experiments_skipped",
                "datasets_skipped",
                "traces_skipped",
                "prompts_skipped",
                "projects_skipped",
            ]
        ]
    )
    total_errors = sum(
        [
            stats.get(key, 0)
            for key in [
                "experiments_errors",
                "datasets_errors",
                "traces_errors",
                "prompts_errors",
                "projects_errors",
            ]
        ]
    )

    table.add_row("", "", "", "", style="bold")
    table.add_row(
        "📦 Total",
        str(total_imported),
        str(total_skipped) if total_skipped > 0 else "",
        str(total_errors) if total_errors > 0 else "",
        style="bold green",
    )

    console.print()
    console.print(table)
    console.print()



"""
Client patchers for CrewAI LLM providers (v1.0.0+).

This module handles patching of LLM clients used by CrewAI agents with Opik tracking.
Each provider has its own patching function that handles missing dependencies gracefully.
"""

import logging
from typing import Any, Optional, TYPE_CHECKING

import crewai

if TYPE_CHECKING:
    import crewai.llms.providers.openai.completion as openai_completion
    import crewai.llms.providers.anthropic.completion as anthropic_completion
    import crewai.llms.providers.gemini.completion as gemini_completion
    import crewai.llms.providers.bedrock.completion as bedrock_completion

LOGGER = logging.getLogger(__name__)


def patch_llm_client(crew: crewai.Crew, project_name: Optional[str]) -> None:
    """
    Patches LLM clients used by CrewAI agents with Opik tracking.

    Handles missing provider libraries gracefully by logging warnings instead of failing.

    Args:
        crew: The Crew instance containing agents to patch.
        project_name: The name of the project to associate with tracking.
    """
    for agent in crew.agents:
        _patch_single_llm_client(agent.llm, project_name)


def _patch_single_llm_client(llm: Any, project_name: Optional[str]) -> None:
    """
    Patches an LLM client based on its provider type.

    Args:
        llm: The CrewAI LLM instance to patch.
        project_name: The name of the project to associate with tracking.
    """
    if _is_openai_llm(llm):
        _patch_openai_client(llm, project_name)
    elif _is_anthropic_llm(llm):
        _patch_anthropic_client(llm, project_name)
    elif _is_gemini_llm(llm):
        _patch_gemini_client(llm, project_name)
    elif _is_bedrock_llm(llm):
        _patch_bedrock_client(llm, project_name)


def _is_openai_llm(llm: Any) -> bool:
    """
    Checks if LLM is an OpenAI provider.

    Args:
        llm: The CrewAI LLM to check.

    Returns:
        True if LLM is OpenAI provider, False otherwise.
    """
    try:
        import crewai.llms.providers.openai.completion

        return isinstance(llm, crewai.llms.providers.openai.completion.OpenAICompletion)
    except ImportError:
        return False


def _is_anthropic_llm(llm: Any) -> bool:
    """
    Checks if LLM is an Anthropic provider.

    Args:
        llm: The CrewAI LLM to check.

    Returns:
        True if LLM is Anthropic provider, False otherwise.
    """
    try:
        import crewai.llms.providers.anthropic.completion

        return isinstance(
            llm, crewai.llms.providers.anthropic.completion.AnthropicCompletion
        )
    except ImportError:
        return False


def _is_gemini_llm(llm: Any) -> bool:
    """
    Checks if LLM is a Gemini provider.

    Args:
        llm: The CrewAI LLM to check.

    Returns:
        True if LLM is Gemini provider, False otherwise.
    """
    try:
        import crewai.llms.providers.gemini.completion

        return isinstance(llm, crewai.llms.providers.gemini.completion.GeminiCompletion)
    except ImportError:
        return False


def _is_bedrock_llm(llm: Any) -> bool:
    """
    Checks if LLM is a Bedrock provider.

    Args:
        llm: The CrewAI LLM to check.

    Returns:
        True if LLM is Bedrock provider, False otherwise.
    """
    try:
        import crewai.llms.providers.bedrock.completion

        return isinstance(
            llm, crewai.llms.providers.bedrock.completion.BedrockCompletion
        )
    except ImportError:
        return False


def _patch_openai_client(
    llm: "openai_completion.OpenAICompletion", project_name: Optional[str]
) -> None:
    """
    Patches OpenAI client for the given LLM.

    Args:
        llm: The CrewAI LLM instance with OpenAI client to patch.
        project_name: The name of the project to associate with tracking.
    """
    try:
        import opik.integrations.openai

        llm.client = opik.integrations.openai.track_openai(
            llm.client, project_name=project_name
        )
    except Exception:
        LOGGER.warning("Failed to track OpenAI client for LLM", exc_info=True)


def _patch_anthropic_client(
    llm: "anthropic_completion.AnthropicCompletion", project_name: Optional[str]
) -> None:
    """
    Patches Anthropic client for the given LLM.

    Args:
        llm: The CrewAI LLM instance with Anthropic client to patch.
        project_name: The name of the project to associate with tracking.
    """
    try:
        import opik.integrations.anthropic

        llm.client = opik.integrations.anthropic.track_anthropic(
            llm.client, project_name=project_name
        )
    except Exception:
        LOGGER.warning("Failed to track Anthropic client for LLM", exc_info=True)


def _patch_gemini_client(
    llm: "gemini_completion.GeminiCompletion", project_name: Optional[str]
) -> None:
    """
    Patches Gemini client for the given LLM.

    Args:
        llm: The CrewAI LLM instance with Gemini client to patch.
        project_name: The name of the project to associate with tracking.
    """
    try:
        import opik.integrations.genai

        llm.client = opik.integrations.genai.track_genai(
            llm.client, project_name=project_name
        )
    except Exception:
        LOGGER.warning("Failed to track Gemini client for LLM", exc_info=True)


def _patch_bedrock_client(
    llm: "bedrock_completion.BedrockCompletion", project_name: Optional[str]
) -> None:
    """
    Patches Bedrock client for the given LLM.

    Args:
        llm: The CrewAI LLM instance with Bedrock client to patch.
        project_name: The name of the project to associate with tracking.
    """
    try:
        import opik.integrations.bedrock

        llm.client = opik.integrations.bedrock.track_bedrock(
            llm.client, project_name=project_name
        )
    except Exception:
        LOGGER.warning("Failed to track Bedrock client for LLM", exc_info=True)



"""
Experiment service for Harbor integration.

This module manages the connection between Harbor benchmark runs and Opik experiments,
enabling evaluation tracking and result visualization.

Harbor Terminology Mapping to Opik:
-----------------------------------
- **Harbor Job**: A benchmark run that evaluates one or more agents on a dataset.
  Maps to an Opik Experiment.

- **Harbor Trial**: A single agent run on a single task within a job.
  Each trial produces one Opik Trace (capturing the agent's execution).

- **Harbor Source**: The benchmark dataset being used (e.g., "terminal-bench", "swe-bench").
  Maps to an Opik Dataset. Each source gets its own dataset.

- **Harbor Task**: A specific problem/challenge within a dataset (e.g., "fix-git" task).
  Maps to an Opik Dataset Item.

Flow Overview:
--------------
1. When a Harbor job starts, this service is initialized with an experiment name.
2. For each unique source (benchmark dataset), we create/get an Opik Dataset and Experiment.
3. For each trial (agent run on a task), we:
   a. Create a dataset item for the task (or reuse existing one if task was run before)
   b. Link the trial's trace to the experiment via ExperimentItemReferences
4. This allows viewing all trial results in Opik's experiment comparison UI.
"""

import logging
from datetime import datetime
from typing import Any, Dict, Optional, Set, TYPE_CHECKING

from opik.api_objects import opik_client
from opik.api_objects.experiment import experiment_item

if TYPE_CHECKING:
    from opik.api_objects.experiment.experiment import Experiment
    from opik.api_objects.dataset.dataset import Dataset

LOGGER = logging.getLogger(__name__)

# Global singleton service instance (one per Harbor job)
_SERVICE: Optional["HarborExperimentService"] = None


class HarborExperimentService:
    """
    Manages Opik datasets and experiments for Harbor benchmark jobs.

    This service handles the mapping between Harbor's evaluation structure and Opik's
    experiment tracking:

    - Each Harbor source (benchmark dataset) → One Opik Dataset + One Opik Experiment
    - Each Harbor task → One Opik Dataset Item
    - Each Harbor trial → One Opik Trace, linked to the experiment

    The service uses lazy initialization - datasets and experiments are created
    on-demand when the first trial for a source is linked.

    Attributes:
        _experiment_name: Name for experiments created by this service.
        _experiment_config: Config dict stored on experiments (agent/model info).
        _client: Cached Opik client instance.
        _datasets: Map of source name → Opik Dataset.
        _experiments: Map of source name → Opik Experiment.
        _linked_trials: Set of trial names already linked (prevents duplicates).
    """

    def __init__(
        self,
        experiment_name: str,
        experiment_config: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Initialize the experiment service.

        Args:
            experiment_name: Name for experiments. Typically includes job_id
                for uniqueness (e.g., "harbor-job-abc123").
            experiment_config: Optional config dict to store on experiments.
                Typically contains agent/model info (e.g., {"agent_name": "terminus",
                "model_name": "gpt-4o"}).
        """
        self._experiment_name = experiment_name
        self._experiment_config = experiment_config or {}
        self._experiment_config["created_from"] = "harbor"
        self._client = opik_client.get_client_cached()

        # Lazy-initialized per source (benchmark dataset)
        self._datasets: Dict[str, "Dataset"] = {}
        self._experiments: Dict[str, "Experiment"] = {}

        # Track which trials have been linked to avoid duplicates
        self._linked_trials: Set[str] = set()

    def _ensure_dataset_and_experiment(self, source: str) -> None:
        """
        Ensure an Opik Dataset and Experiment exist for the given source.

        Creates them lazily on first access. Each Harbor source (benchmark dataset)
        gets its own Opik Dataset and Experiment pair.

        Args:
            source: The Harbor source/benchmark name (e.g., "terminal-bench").
        """
        if source in self._experiments:
            return

        try:
            # Create or get the dataset for this benchmark source
            dataset = self._client.get_or_create_dataset(
                name=source,
                description=f"Harbor benchmark dataset: {source}",
            )
            self._datasets[source] = dataset
            LOGGER.info("Using dataset '%s' for Harbor source", source)

            # Create a new experiment for this job run
            experiment = self._client.create_experiment(
                dataset_name=source,
                name=self._experiment_name,
                experiment_config=self._experiment_config,
            )
            self._experiments[source] = experiment
            LOGGER.info(
                "Created experiment '%s' for dataset '%s'",
                self._experiment_name,
                source,
            )
        except Exception as e:
            LOGGER.warning(
                "Failed to create dataset/experiment for source '%s': %s",
                source,
                e,
            )

    def link_trial_to_experiment(
        self,
        trial_name: str,
        trace_id: str,
        source: Optional[str] = None,
        task_name: Optional[str] = None,
    ) -> None:
        """
        Link a Harbor trial's trace to the Opik experiment.

        This creates the connection between a trial's execution trace and the
        experiment, enabling the trial to appear in Opik's experiment comparison UI.

        The flow:
        1. Ensure dataset and experiment exist for the source
        2. Create or find the dataset item for this task
        3. Link the trace to the experiment via the dataset item

        Args:
            trial_name: Unique identifier for the trial (e.g., "task__abc123").
                Used to prevent duplicate linking.
            trace_id: The Opik trace ID for this trial's execution.
            source: The Harbor source/benchmark name. Defaults to "harbor-default".
            task_name: The task name within the benchmark (e.g., "fix-git").
                Used to create/find the dataset item.
        """
        source = source or "harbor-default"

        # Prevent duplicate linking of the same trial
        if trial_name in self._linked_trials:
            return

        # Ensure we have a dataset and experiment for this source
        self._ensure_dataset_and_experiment(source)

        experiment = self._experiments.get(source)
        dataset = self._datasets.get(source)

        if experiment is None or dataset is None:
            LOGGER.warning(
                "Failed to create experiment/dataset for source '%s', "
                "trial '%s' will not be linked",
                source,
                trial_name,
            )
            return

        try:
            # Insert the task as a dataset item (idempotent - duplicates are handled)
            dataset.insert([{"task_name": task_name}])

            # Find the dataset item ID for this task.
            # We search because the same task may have been inserted in a previous run,
            # and we want to reuse the existing item ID for proper experiment linking.
            items = dataset.get_items()
            dataset_item_id = None
            for item in items:
                if item.get("task_name") == task_name:
                    dataset_item_id = item.get("id")
                    break

            if dataset_item_id is None:
                LOGGER.warning("Could not find dataset item for task '%s'", task_name)
                return

            # Link the trace to the experiment via the dataset item
            experiment.insert(
                [
                    experiment_item.ExperimentItemReferences(
                        dataset_item_id=dataset_item_id,
                        trace_id=trace_id,
                    )
                ]
            )

            self._linked_trials.add(trial_name)
            LOGGER.debug(
                "Linked trial '%s' (trace %s) to experiment '%s'",
                trial_name,
                trace_id,
                self._experiment_name,
            )
        except Exception as e:
            LOGGER.warning("Failed to link trial '%s' to experiment: %s", trial_name, e)


def setup_lazy(
    experiment_name: Optional[str] = None,
    experiment_config: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Setup the experiment service lazily.

    Called when the first Harbor trial runs. Creates the global service instance
    that will be used for all subsequent trial linking. Datasets and experiments
    are created on-demand when trials are linked.

    Args:
        experiment_name: Name for the experiment. If None, auto-generates
            a timestamped name like "harbor-20241209-143000".
        experiment_config: Optional config dict to store on experiments.
            Typically contains agent/model info (e.g., {"agent_name": "terminus",
            "model_name": "gpt-4o"}).
    """
    global _SERVICE

    if _SERVICE is not None:
        LOGGER.debug("Experiment service already setup, skipping")
        return

    if experiment_name is None:
        timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        experiment_name = f"harbor-{timestamp}"

    _SERVICE = HarborExperimentService(
        experiment_name=experiment_name,
        experiment_config=experiment_config,
    )

    LOGGER.info("Experiment service setup for '%s'", experiment_name)


def get_service() -> Optional[HarborExperimentService]:
    """Get the current experiment service instance, or None if not initialized."""
    return _SERVICE


def reset() -> None:
    """Reset the experiment service. Used for testing."""
    global _SERVICE
    _SERVICE = None



import dataclasses
import logging
import math
import time
from concurrent.futures import Future, CancelledError
from typing import Callable, List, Optional

import httpx

from . import (
    base_upload_manager,
    upload_options,
    file_upload_monitor,
    thread_pool,
    file_uploader,
)
from .. import format_helpers, synchronization
from ..message_processing import messages
from ..rest_api import client as rest_api_client

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class UploadResult:
    future: Future
    monitor: file_upload_monitor.FileUploadMonitor
    upload_options: upload_options.FileUploadOptions

    def ready(self) -> bool:
        """Allows to check if wrapped Future successfully finished"""
        return self.future.done()

    def successful(self, timeout: Optional[float] = None) -> bool:
        """Allows to check if wrapped Future completed without raising an exception"""
        try:
            exception = self.future.exception(timeout)
            if exception is None:
                return True

            raise exception
        except (CancelledError, TimeoutError) as e:
            LOGGER.warning(
                "Timeout while waiting for the result of file '%s' upload. Error: %s",
                self.upload_options.file_name,
                e,
            )
        except Exception as exception:
            LOGGER.error(
                "Failed to upload file with name '%s' from path [%s] with size [%s]. Error: %s",
                self.upload_options.file_name,
                self.upload_options.file_path,
                format_helpers.format_bytes(self.upload_options.file_size),
                exception,
                exc_info=True,
            )

        return False


class FileUploadManagerMonitor:
    def __init__(
        self, file_upload_manager: base_upload_manager.BaseFileUploadManager
    ) -> None:
        self.file_upload_manager = file_upload_manager
        self.last_remaining_bytes = 0
        self.last_remaining_uploads_display: Optional[float] = None

    def log_remaining_uploads(self) -> None:
        remaining = self.file_upload_manager.remaining_data()

        current_time = time.monotonic()

        if remaining.bytes == 0:
            LOGGER.info(
                "All assets have been sent, waiting for delivery confirmation",
            )
        elif self.last_remaining_uploads_display is None:
            LOGGER.info(
                "Still uploading %d file(s), remaining %s/%s",
                remaining.uploads,
                format_helpers.format_bytes(remaining.bytes),
                format_helpers.format_bytes(remaining.total_size),
            )
        else:
            uploaded_bytes = self.last_remaining_bytes - remaining.bytes
            time_elapsed = current_time - self.last_remaining_uploads_display
            upload_speed = uploaded_bytes / time_elapsed

            # Avoid 0 division if no bytes were uploaded in the last period
            if uploaded_bytes <= 0:
                # avoid negative upload speed
                if upload_speed < 0:
                    upload_speed = 0

                LOGGER.info(
                    "Still uploading %d file(s), remaining %s/%s, Throughput %s/s, ETA unknown",
                    remaining.uploads,
                    format_helpers.format_bytes(remaining.bytes),
                    format_helpers.format_bytes(remaining.total_size),
                    format_helpers.format_bytes(upload_speed),
                )

            else:
                remaining_time = str(int(math.ceil(remaining.bytes / upload_speed)))

                LOGGER.info(
                    "Still uploading %d file(s), remaining %s/%s, Throughput %s/s, ETA ~%ss",
                    remaining.uploads,
                    format_helpers.format_bytes(remaining.bytes),
                    format_helpers.format_bytes(remaining.total_size),
                    format_helpers.format_bytes(upload_speed),
                    remaining_time,
                )

        self.last_remaining_bytes = remaining.bytes
        self.last_remaining_uploads_display = current_time

    def all_done(self) -> bool:
        return self.file_upload_manager.all_done()


class FileUploadManager(base_upload_manager.BaseFileUploadManager):
    """Manages concurrent file uploads."""

    def __init__(
        self,
        rest_client: rest_api_client.OpikApi,
        httpx_client: httpx.Client,
        worker_count: int,
    ) -> None:
        self._httpx_client = httpx_client
        self._rest_client = rest_client

        self._executor = thread_pool.get_thread_pool(worker_count=worker_count)
        self._upload_results: List[UploadResult] = []
        self.closed = False

    def upload(self, message: messages.BaseMessage) -> None:
        if isinstance(message, messages.CreateAttachmentMessage):
            self.upload_attachment(message)
        else:
            raise ValueError(f"Message {message} is not supported for file upload.")

    def upload_attachment(self, attachment: messages.CreateAttachmentMessage) -> None:
        assert isinstance(attachment, messages.CreateAttachmentMessage), (
            "Wrong attachment message type"
        )

        options = upload_options.file_upload_options_from_attachment(attachment)
        self._submit_upload(
            uploader=file_uploader.upload_attachment,
            options=options,
        )

    def _submit_upload(
        self, options: upload_options.FileUploadOptions, uploader: Callable
    ) -> None:
        if self.closed:
            LOGGER.warning(
                "The file upload manager has been already closed. No more files can be submitted for upload. (%s)",
                options.file_name,
            )
            return

        monitor = file_upload_monitor.FileUploadMonitor()
        if options.file_size > 0:
            monitor.total_size = options.file_size

        kwargs = {
            "monitor": monitor,
            "upload_options": options,
            "rest_client": self._rest_client,
            "upload_httpx_client": self._httpx_client,
        }
        future = self._executor.submit(uploader, **kwargs)
        self._upload_results.append(
            UploadResult(future, monitor=monitor, upload_options=options)
        )

    def all_done(self) -> bool:
        return all(result.ready() for result in self._upload_results)

    def remaining_data(self) -> base_upload_manager.RemainingUploadData:
        remaining_uploads = 0
        remaining_bytes_to_upload = 0
        total_size = 0
        for result in self._upload_results:
            if result.ready() is True:
                continue

            remaining_uploads += 1

            monitor = result.monitor
            if monitor.total_size is None or monitor.bytes_sent is None:
                continue

            total_size += monitor.total_size
            remaining_bytes_to_upload += monitor.total_size - monitor.bytes_sent

        return base_upload_manager.RemainingUploadData(
            uploads=remaining_uploads,
            bytes=remaining_bytes_to_upload,
            total_size=total_size,
        )

    def remaining_uploads(self) -> int:
        """Returns the number of remaining uploads. Non-blocking - can be called at any time."""
        status_list = [result.ready() for result in self._upload_results]
        return status_list.count(False)

    def failed_uploads(self, timeout: Optional[float]) -> int:
        """Important - this is blocking method waiting for all remaining uploads to complete or while
        timeout is expired."""
        failed = 0
        for result in self._upload_results:
            if not result.ready() or not result.successful(timeout):
                failed += 1

        return failed

    def flush(self, timeout: Optional[float], sleep_time: int = 5) -> bool:
        """Flushes all pending uploads. This is a blocking method that waits for all remaining uploads to complete,
        either until they finish or the specified timeout expires. If no timeout is set, it waits indefinitely.
        Args:
             timeout: Timeout in seconds to wait for all remaining uploads to complete.
                If None is provided, it will wait for all remaining uploads to complete.
            sleep_time: The sleep interval between checks and printing progress.
        Returns:
            The flag to indicate whether all remaining uploads are completed or not within the provided timeout.
        """
        upload_monitor = FileUploadManagerMonitor(self)

        synchronization.wait_for_done(
            check_function=lambda: self.all_done(),
            progress_callback=upload_monitor.log_remaining_uploads,
            timeout=timeout,
            sleep_time=sleep_time,
        )

        # check failed uploads number only if all upload operations completed to avoid blocking
        if self.all_done():
            failed_uploads = self.failed_uploads(timeout)
            if failed_uploads > 0:
                LOGGER.warning(
                    "Failed to upload %d file(s). Check logs for details.",
                    failed_uploads,
                )
            return True

        return False

    def close(self) -> None:
        self._executor.shutdown(wait=True)
        self.closed = True
